__author__ = "Simon Nilsson"

import glob
import os
from datetime import datetime
from typing import Dict, Union

import pandas as pd
import seaborn as sns

from simba.mixins.config_reader import ConfigReader
from simba.utils.checks import check_if_filepath_list_is_empty
from simba.utils.data import detect_bouts
from simba.utils.enums import ConfigKey, Dtypes, Paths
from simba.utils.read_write import (get_fn_ext, read_config_entry,
                                    read_config_file, read_df,
                                    read_project_path_and_file_type,
                                    read_video_info_csv)


class PupRetrieverCalculator(ConfigReader):
    """
    Pup retreival calculator used in ``Winters et al., `Sci Reports`, 2022``

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter dict settings: user-defined setting for pup retrieval.

    .. note::
       `Documentation <https://github.com/sgoldenlab/simba/blob/master/docs/add_on_pup_ret.md>`_

    Examples
    ----------
    >>> settings = {'pup_track_p': 0.025, 'dam_track_p': 0.5, 'start_distance_criterion': 80.0, 'carry_frames': 90.0, 'core_nest': 'corenest', 'nest': 'nest', 'dam_name': '1_mother', 'pup_name': '2_pup', 'smooth_function': 'gaussian', 'smooth_factor': 5, 'max_time': 90.0, 'clf_carry': 'carry', 'clf_approach': 'approach', 'clf_dig': 'digging', 'distance_plots': True, 'log': True, 'swarm_plot': True}
    >>> config_path = '/Users/simon/Downloads/Automated PRT_test/project_folder/project_config.ini'
    >>> calculator = PupRetrieverCalculator(config_path=config_path, settings=settings)
    >>> calculator.run()

    References
    ----------
    .. [1] Winters et al., Automated procedure to assess pup retrieval in laboratory mice, `Sci Reports`, 2022.
    """

    def __init__(self, config_path: str, settings: Dict[str, Union[float, str, bool]]):
        ConfigReader.__init__(config_path=config_path)
        self.config = read_config_file(config_path=config_path)
        self.project_path, self.file_type = read_project_path_and_file_type(
            config=self.config
        )
        self.settings, self.datetime = settings, datetime.now().strftime("%Y%m%d%H%M%S")
        self.animal_cnt = read_config_entry(
            self.config,
            ConfigKey.GENERAL_SETTINGS.value,
            ConfigKey.ANIMAL_CNT.value,
            Dtypes.INT.value,
        )
        self.clf_lst = [
            settings["clf_approach"],
            settings["clf_carry"],
            settings["clf_dig"],
        ]
        self.distance_pup_core_field = (
            f'{self.settings["core_nest"]} {self.settings["pup_name"]} {"distance"}'
        )
        self.distance_dam_core_field = (
            f'{self.settings["core_nest"]} {self.settings["dam_name"]} {"distance"}'
        )
        self.pup_in_core_field = (
            f'{self.settings["core_nest"]} {self.settings["pup_name"]} {"in zone"}'
        )
        self.pup_in_nest_field = (
            f'{self.settings["nest"]} {self.settings["pup_name"]} {"in zone"}'
        )

        machine_results_path = os.path.join(
            self.project_path, Paths.MACHINE_RESULTS_DIR.value
        )
        self.logs_dir_path = os.path.join(self.project_path, "logs")

        self.data_files = glob.glob(machine_results_path + "/*." + self.file_type)
        check_if_filepath_list_is_empty(
            filepaths=self.data_files,
            error_msg="SIMBA ERROR: NO FILES FOUND IN {}".format(machine_results_path),
        )
        self.vid_info_df = read_video_info_csv(
            os.path.join(self.project_path, Paths.VIDEO_INFO.value)
        )

    def __get_max_frames(self):
        self.max_frames = int(self.fps * self.settings["max_time"])
        if self.max_frames < len(self.data_df):
            self.data_df = self.data_df.head(self.max_frames)

    def __check_column_names(self):
        for c in [
            self.distance_pup_core_field,
            self.distance_dam_core_field,
            self.pup_in_core_field,
            self.pup_in_nest_field,
        ]:
            if c not in self.data_df.columns:
                print(
                    f"SIMBA ERROR: Could not find anticipated column named {c} in {self.file_path}"
                )
                raise ValueError()

    def correct_in_nest_frames(self):
        if 1 in self.data_df[self.pup_in_nest_field].values:
            for nest_frame in self.data_df[
                self.data_df[self.pup_in_nest_field] == 1
            ].index.tolist():
                sliced_df = (
                    self.data_df[self.settings["clf_carry"]]
                    .loc[nest_frame - self.carry_frames + 1 : nest_frame]
                    .tolist()
                )
                if sum(sliced_df) == 0:
                    self.data_df.at[nest_frame, self.pup_in_nest_field] = 0
                else:
                    break

    def __create_log(self):
        log = {
            "Datetime": self.datetime,
            "Videos_#": len(self.data_files),
            **self.settings,
        }
        log_df = pd.DataFrame.from_dict(log, orient="index").rename(
            columns={0: "VALUES"}
        )
        log_save_path = os.path.join(
            self.logs_dir_path, f"Log_pup_retrieval_{self.datetime}"
        )
        log_df.to_csv(log_save_path)
        print("Pup retreival log saved at {}...".format(log_save_path))

    def __generate_figure(
        self,
        data: pd.DataFrame,
        y_col: str,
        x_lbl: str,
        y_lbl: str,
        title: str,
        hue: str,
        video_name: str,
    ):
        current_figure = sns.scatterplot(
            x=data.index, y=data[y_col], hue=data[hue], legend=False, palette="Set1"
        )
        current_figure.set(xlabel=x_lbl, ylabel=y_lbl, title=title)
        save_plot_name = f"{title} {video_name} {self.datetime}.png"
        save_plot_path = os.path.join(self.logs_dir_path, video_name)
        if not os.path.exists(save_plot_path):
            os.makedirs(save_plot_path)
        image_save_path = os.path.join(save_plot_path, save_plot_name)
        current_figure.figure.savefig(image_save_path, bbox_inches="tight")
        current_figure.clear()

    def __create_swarm_plot(self):
        figure_df = self.out_df.copy()
        figure_df["Experiment"] = 1
        swarm_plot = sns.swarmplot(
            x="Experiment", y="PUP IN NEST (S)", data=figure_df, color="grey"
        )
        swarm_plot.set(
            xlabel="",
            ylabel="Pup in nest (s)",
            title="Summary - pup retrieval time (s)",
        )
        swarm_plot_name = f"Summary_pup_retrieval_times_{self.datetime}.png"
        save_plot_path = os.path.join(self.logs_dir_path, swarm_plot_name)
        swarm_plot.figure.savefig(save_plot_path, bbox_inches="tight")
        swarm_plot.clear()
        print(f"Swarm plot saved @ {save_plot_path}...")

    def run(self):
        self.out = []
        for file_cnt, file_path in enumerate(self.data_files):
            self.results = {}
            self.file_path = file_path
            _, video_name, _ = get_fn_ext(filepath=file_path)
            _, _, self.fps = self.read_video_info(video_name=video_name)
            self.data_df = read_df(
                file_path=file_path, file_type=self.file_type
            ).fillna(method="ffill")
            self.__get_max_frames()
            self.carry_frames = int(self.fps * self.settings["carry_time"])

            self.data_df["mean_p_mother"] = self.data_df[
                self.animal_bp_dict[self.settings["dam_name"]]["P_bps"]
            ].mean(axis=1)
            self.data_df["pup_p_mean"] = self.data_df[
                self.animal_bp_dict[self.settings["pup_name"]]["P_bps"]
            ].mean(axis=1)
            self.data_df["cumsum_nest_pup"] = self.data_df[
                self.pup_in_nest_field
            ].cumsum()

            if self.settings["distance_plots"]:
                self.__generate_figure(
                    data=self.data_df,
                    y_col=self.distance_dam_core_field,
                    x_lbl="frame number",
                    y_lbl="distance (mm)",
                    title="distance between mother and corenest - BEFORE pre-processing",
                    hue="cumsum_nest_pup",
                    video_name=video_name,
                )

                self.__generate_figure(
                    data=self.data_df,
                    y_col=self.distance_pup_core_field,
                    x_lbl="frame number",
                    y_lbl="distance (mm)",
                    title="distance between pup and corenest - BEFORE pre-processing",
                    hue="cumsum_nest_pup",
                    video_name=video_name,
                )

            for clf in self.clf_lst:
                self.data_df.loc[
                    self.data_df["mean_p_mother"] < self.settings["dam_track_p"], clf
                ] = 0

            first_row = self.data_df[
                self.data_df[self.distance_pup_core_field]
                > self.settings["start_distance_criterion"]
            ].index[0]
            self.data_df.loc[0:first_row, self.distance_pup_core_field] = (
                self.data_df.loc[first_row, self.distance_pup_core_field]
            )
            self.data_df.loc[0:first_row, self.pup_in_core_field] = 0
            self.data_df.loc[0:first_row, self.pup_in_nest_field] = 0
            self.correct_in_nest_frames()

            rows_with_low_mean_pup_prob = self.data_df[
                self.data_df["pup_p_mean"] < self.settings["pup_track_p"]
            ].index.tolist()
            self.data_df.loc[rows_with_low_mean_pup_prob, self.pup_in_core_field] = 0
            self.data_df.loc[rows_with_low_mean_pup_prob, self.pup_in_nest_field] = 0

            if self.settings["smooth_function"] == "gaussian":
                self.data_df[self.distance_pup_core_field] = (
                    self.data_df[self.distance_pup_core_field]
                    .rolling(window=int(self.fps), win_type="gaussian", center=True)
                    .mean(std=self.settings["smooth_factor"])
                    .fillna(self.data_df[self.distance_pup_core_field])
                )
                self.data_df[self.distance_dam_core_field] = (
                    self.data_df[self.distance_dam_core_field]
                    .rolling(window=int(self.fps), win_type="gaussian", center=True)
                    .mean(std=self.settings["smooth_factor"])
                    .fillna(self.data_df[self.distance_dam_core_field])
                )

            if self.settings["distance_plots"]:
                self.__generate_figure(
                    data=self.data_df,
                    y_col=self.distance_dam_core_field,
                    x_lbl="frame number",
                    y_lbl="distance (mm)",
                    title="distance between mother and corenest - AFTER pre-processing",
                    hue="cumsum_nest_pup",
                    video_name=video_name,
                )

                self.__generate_figure(
                    data=self.data_df,
                    y_col=self.distance_pup_core_field,
                    x_lbl="frame number",
                    y_lbl="distance (mm)",
                    title="distance between pup and core-nest - AFTER pre-processing",
                    hue="cumsum_nest_pup",
                    video_name=video_name,
                )

            closest_dist_between_pup_and_zone = round(
                self.data_df[self.distance_pup_core_field].min(), 3
            )
            if 1 in self.data_df[self.pup_in_nest_field].values:
                frame_when_pup_is_in_zone = self.data_df[
                    self.data_df[self.pup_in_nest_field] == 1
                ].index.min()
                time_seconds_until_zone = round(frame_when_pup_is_in_zone / self.fps, 3)
                reason_zone = "Pup in nest"
            else:
                frame_when_pup_is_in_zone = len(self.data_df)
                time_seconds_until_zone = round(frame_when_pup_is_in_zone / self.fps, 3)
                reason_zone = "Pup not retrieved"

            if 1 in self.data_df[self.pup_in_core_field].values:
                frame_when_pup_is_in_core_nest = self.data_df[
                    self.data_df[self.pup_in_core_field] == 1
                ].index.min()
                time_seconds_until_corenest = round(
                    frame_when_pup_is_in_core_nest / self.fps, 3
                )
                reason_corenest = "Pup in core-nest"
            else:
                frame_when_pup_is_in_core_nest = len(self.data_df)
                time_seconds_until_corenest = round(
                    frame_when_pup_is_in_core_nest / self.fps, 3
                )
                reason_corenest = "Pup not in core-nest"

            latencies, total_times, before_retrieval_time = {}, {}, {}
            for clf in self.clf_lst:
                total_times[clf] = round(self.data_df[clf].sum() / self.fps, 3)
                before_retrieval_time[clf] = round(
                    self.data_df.loc[0:frame_when_pup_is_in_zone, clf].sum() / self.fps,
                    3,
                )
                latencies[clf] = round(
                    self.data_df[self.data_df[clf] == 1].index.min() / self.fps, 3
                )

            event_counter, time_between_events, mean_duration = {}, {}, {}
            bouts = detect_bouts(
                data_df=self.data_df, target_lst=self.clf_lst, fps=self.fps
            )
            for clf in self.clf_lst:
                clf_bouts = bouts[bouts["Event"] == clf].reset_index(drop=True)
                event_counter[clf] = len(clf_bouts)
                mean_duration[clf] = round(clf_bouts["Bout_time"].mean() / self.fps, 3)
                clf_bouts["Start_time"] = clf_bouts["Start_time"].shift(-1)
                clf_bouts.drop(clf_bouts.tail(1).index, inplace=True)
                clf_bouts["TIME BETWEEN"] = (
                    clf_bouts["Start_time"] - clf_bouts["End Time"]
                )
                time_between_events[clf] = round(
                    clf_bouts["TIME BETWEEN"].mean() / self.fps, 3
                )

            before_enter_core_df = self.data_df.loc[
                0 : frame_when_pup_is_in_core_nest - 1
            ]
            before_core_bouts = detect_bouts(
                data_df=before_enter_core_df, target_lst=self.clf_lst, fps=self.fps
            )
            event_counter_before_corenest, mean_event_length_before_corenest = {}, {}
            for clf in self.clf_lst:
                clf_bouts = before_core_bouts[
                    before_core_bouts["Event"] == clf
                ].reset_index(drop=True)
                mean_event_length_before_corenest[clf] = round(
                    clf_bouts["Bout_time"].mean(), 3
                )
                event_counter_before_corenest[clf] = len(clf_bouts)

            retrieval_frame = self.data_df[
                self.data_df[self.pup_in_nest_field] == 1
            ].index.min()
            retrieval_frame = 160
            first_approach = self.data_df[
                self.data_df[self.settings["clf_approach"]] == 1
            ].index.min()
            dig_bouts = detect_bouts(
                data_df=before_enter_core_df,
                target_lst=[self.settings["clf_dig"]],
                fps=self.fps,
            )
            dig_bouts_in_window = dig_bouts.loc[
                (dig_bouts["End_frame"] > first_approach)
                & (dig_bouts["End_frame"] < retrieval_frame)
            ]
            dig_bouts_in_window_seconds = round(
                dig_bouts_in_window["Bout_time"].sum(), 3
            )

            self.results["VIDEO"] = video_name
            self.results["PUP IN NEST (FRAME)"] = frame_when_pup_is_in_zone
            self.results["PUP IN NEST (S)"] = time_seconds_until_zone
            self.results["MINIMUM DISTANCE (PUP TO CORENEST)"] = (
                closest_dist_between_pup_and_zone
            )
            self.results["REASON (PUP IN NEST)"] = reason_zone
            self.results["PUP IN CORE-NEST (FRAME)"] = frame_when_pup_is_in_core_nest
            self.results["PUP IN CORE-NEST (S)"] = time_seconds_until_corenest
            self.results["REASON (PUP IN CORE-NEST)"] = reason_corenest

            for clf in self.clf_lst:
                self.results[clf + " (TOTAL TIME)"] = total_times[clf]
                self.results[clf + " (BEFORE RETRIEVAL)"] = (
                    event_counter_before_corenest[clf]
                )
                self.results[clf + " (LATENCY TO FIRST EVENT)"] = latencies[clf]
                self.results[clf + " (EVENT COUNT)"] = event_counter[clf]
                self.results[clf + " (MEAN DURATION)"] = mean_duration[clf]
                self.results[clf + " (MEAN INTERVAL)"] = time_between_events[clf]

            self.results["DIG TIME AFTER APPROACH AND BEFORE RETRIEVAL (S)"] = (
                dig_bouts_in_window_seconds
            )
            self.results["DIG EVENTS AFTER APPROACH"] = len(dig_bouts_in_window)
            self.results["MEAN DIG DURATION AFTER APPROACH"] = round(
                dig_bouts_in_window["Bout_time"].mean(), 3
            )
            self.out.append(pd.DataFrame.from_dict(self.results, orient="index").T)

        self.out_df = pd.concat(self.out, axis=0).reset_index(drop=True)

        if self.settings["log"]:
            self.__create_log()

        if self.settings["swarm_plot"]:
            self.__create_swarm_plot()

    def save_results(self):
        file_path = os.path.join(
            self.logs_dir_path, f"Log_pup_retrieval_{self.datetime}.csv"
        )
        self.out_df.to_csv(file_path)
        print(f"SIMBA COMPLETE: Summary data saved at {file_path}.")


# settings = {'pup_track_p': 0.025, 'dam_track_p': 0.5, 'start_distance_criterion': 80.0, 'carry_frames': 90.0,
#      'core_nest': 'corenest', 'nest': 'nest', 'dam_name': '1_mother', 'pup_name': '2_pup',
#      'smooth_function': 'gaussian', 'smooth_factor': 5, 'max_time': 90.0, 'clf_carry': 'carry',
#      'clf_approach': 'approach', 'clf_dig': 'digging', 'distance_plots': True, 'log': True, 'swarm_plot': True}
# config_path = '/Users/simon/Downloads/Automated PRT_test/project_folder/project_config.ini'
#
# test = PupRetrieverCalculator(config_path=config_path, settings=settings)
# test.run()



__author__ = "Simon Nilsson"

import glob
import os
from collections import defaultdict
from statistics import mean
from typing import List, Union

import numpy as np
import pandas as pd

from simba.cue_light_tools.cue_light_tools import find_frames_when_cue_light_on
from simba.mixins.config_reader import ConfigReader
from simba.roi_tools.ROI_analyzer import ROIAnalyzer
from simba.utils.printing import stdout_success
from simba.utils.read_write import get_fn_ext, read_df


class CueLightMovementAnalyzer(ConfigReader):
    """
    Compute aggregate statistics of animal movement in relation to the cue light
    ON and OFF states.

    :parameter str config_path: path to SimBA project config file in Configparser format
    :parameter int pre_window: Time period (in millisecond) before the onset of each cue light to compute aggregate classification
        statistics within.
    :parameter int post_window: Time period (in millisecond) after the offset of each cue light to compute aggregate classification
        statistics within.
    :parameter List[str] cue_light_names: Names of cue lights, as defined in the SimBA ROI interface.
    :parameter float threshold: The body-part post-estimation probability threshold. SimBA omits movement calculations for frames where the
        body-part probability threshold is lower than the user-specified threshold.
    :parameter bool roi_setting: If True, SimBA calculates movement statistics within non-light cue ROIs.

    .. note:
       `Cue light tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/cue_light_tutorial.md>`__.


    Examples
    ----------
    >>> cue_light_movement_analyzer = CueLightMovementAnalyzer(config_path='MyProjectConfig', cue_light_names=['Cue_light'], pre_window=1000, post_window=1000, threshold=0.0, roi_setting=True)
    >>> cue_light_movement_analyzer.calculate_whole_session_movement()
    >>> cue_light_movement_analyzer.organize_results()
    >>> cue_light_movement_analyzer.save_results()
    """

    def __init__(
        self,
        config_path: Union[str, os.PathLike],
        pre_window: int,
        post_window: int,
        cue_light_names: List[str],
        threshold: float,
        roi_setting: bool,
    ):
        ConfigReader.__init__(self, config_path=config_path)
        self.cue_light_names = cue_light_names
        self.roi_setting = roi_setting
        self.p_threshold = threshold
        self.pre_window, self.post_window = pre_window, post_window
        self.in_dir = os.path.join(self.project_path, "csv", "cue_lights")
        self.files_found = glob.glob(self.in_dir + "/*." + self.file_type)
        self.bp_dict, self.bp_columns = defaultdict(list), []
        for cnt, animal in enumerate(self.multi_animal_id_list):
            bp_name = self.config.get("ROI settings", "animal_" + str(cnt + 1) + "_bp")
            if bp_name == "None":
                print("SIMBA ERROR: Please analyze ROI data and set body-parts first.")
                raise ValueError
            for c in ["_x", "_y", "_p"]:
                self.bp_dict[animal].append(bp_name + c)
                self.bp_columns.append(bp_name + c)
        print("Analyzing {} files...".format(str(len(self.files_found))))

        if roi_setting:
            self.roi_analyzer = ROIAnalyzer(
                ini_path=self.config_path,
                data_path=os.path.join(
                    self.project_path, "csv", "outlier_corrected_movement_location"
                ),
                calculate_distances=False,
            )
            self.roi_analyzer.run()
            # self.entries_exits_df = pd.DataFrame.from_dict(self.roi_analyzer.detailed_df, orient='index')
            print(self.roi_analyzer.detailed_df)
            self.entries_exits_df = self.roi_analyzer.detailed_df[
                ~self.roi_analyzer.detailed_df["SHAPE"].isin(self.cue_light_names)
            ]
            if len(self.entries_exits_df) > 0:
                self.entries_exits_df["inside_lst"] = self.entries_exits_df.apply(
                    lambda x: list(
                        range(int(x["ENTRY FRAMES"]), int(x["EXIT FRAMES"] + 1))
                    ),
                    axis=1,
                )

    def __euclidean_distance(
        self, bp_1_x_vals, bp_2_x_vals, bp_1_y_vals, bp_2_y_vals, px_per_mm
    ):
        series = (
            np.sqrt((bp_1_x_vals - bp_2_x_vals) ** 2 + (bp_1_y_vals - bp_2_y_vals) ** 2)
        ) / px_per_mm
        return series

    def calculate_whole_session_movement(self):
        """
        Method to calculate movement in relation to cue-light(s) onsets.

        Returns
        -------
        Attribute: dict
            roi_results
        """

        self.results = {}
        if self.roi_setting:
            self.roi_results = {}
        for file_cnt, file_path in enumerate(self.files_found):
            _, self.video_name, _ = get_fn_ext(file_path)
            self.results[self.video_name] = {}
            self.data_df = read_df(file_path, self.file_type).reset_index(drop=True)
            self.video_info_settings, self.px_per_mm, self.fps = self.read_video_info(
                video_name=self.video_name
            )
            self.prior_window_frames_cnt = int(self.pre_window / (1000 / self.fps))
            self.post_window_frames_cnt = int(self.post_window / (1000 / self.fps))
            self.light_on_dict = find_frames_when_cue_light_on(
                data_df=self.data_df,
                cue_light_names=self.cue_light_names,
                fps=self.fps,
                prior_window_frames_cnt=self.prior_window_frames_cnt,
                post_window_frames_cnt=self.post_window_frames_cnt,
            )

            for animal_name, animal_bps in self.bp_dict.items():
                self.results[self.video_name][animal_name] = {}
                animal_df = self.data_df[animal_bps]
                if self.p_threshold > 0.00:
                    animal_df = animal_df[animal_df[animal_bps[2]] >= self.p_threshold]
                animal_df = animal_df.iloc[:, 0:2].reset_index(drop=True)
                df_shifted = animal_df.shift(1)
                df_shifted = df_shifted.combine_first(animal_df).add_suffix("_shifted")
                animal_df = pd.concat([animal_df, df_shifted], axis=1)
                self.movement = (
                    self.__euclidean_distance(
                        animal_df[animal_bps[0]],
                        animal_df[animal_bps[0] + "_shifted"],
                        animal_df[animal_bps[1]],
                        animal_df[animal_bps[1] + "_shifted"],
                        self.px_per_mm,
                    )
                    / 10
                )
                self.light_movement = {}
                for cue_light in self.cue_light_names:
                    self.light_movement[cue_light] = {}
                    self.light_movement[cue_light]["pre_window_movement"] = (
                        self.movement[
                            self.movement.index.isin(
                                self.light_on_dict[cue_light]["pre_window_frames"]
                            )
                        ]
                    )
                    self.light_movement[cue_light]["light_movement"] = self.movement[
                        self.movement.index.isin(
                            self.light_on_dict[cue_light]["light_on_frames"]
                        )
                    ]
                    self.light_movement[cue_light]["post_window_movement"] = (
                        self.movement[
                            self.movement.index.isin(
                                self.light_on_dict[cue_light]["post_window_frames"]
                            )
                        ]
                    )
                for cue_light in self.cue_light_names:
                    self.results[self.video_name][animal_name][cue_light] = {}
                    for state_name, df in zip(
                        ["pre-cue", "cue", "post-cue"],
                        [
                            self.light_movement[cue_light]["pre_window_movement"],
                            self.light_movement[cue_light]["light_movement"],
                            self.light_movement[cue_light]["post_window_movement"],
                        ],
                    ):
                        self.results[self.video_name][animal_name][cue_light][
                            state_name
                        ] = {}
                        self.results[self.video_name][animal_name][cue_light][
                            state_name
                        ]["Distance (cm)"] = round((df.sum() / 10), 4)
                        velocity_lst = []
                        for sliced_df in np.array_split(df, self.fps):
                            velocity_lst.append(sliced_df.sum())
                        self.results[self.video_name][animal_name][cue_light][
                            state_name
                        ]["Velocity (cm/s)"] = round((mean(velocity_lst) / 10), 4)
                if self.roi_setting:
                    self.roi_results[self.video_name] = {}
                    self.roi_results[self.video_name][animal_name] = {}
                    for roi_name in self.entries_exits_df["SHAPE"].unique():
                        self.roi_results[self.video_name][animal_name][roi_name] = {}
                        inside_roi_frms = self.entries_exits_df["inside_lst"][
                            self.entries_exits_df["SHAPE"] == roi_name
                        ]
                        inside_roi_frms = [i for s in inside_roi_frms for i in s]
                        for cue_light in self.cue_light_names:
                            self.roi_results[self.video_name][animal_name][roi_name][
                                cue_light
                            ] = {}
                            self.overlap_light = list(
                                set(inside_roi_frms).intersection(
                                    self.light_on_dict[cue_light]["light_on_frames"]
                                )
                            )
                            self.overlap_pre_window_frames = list(
                                set(inside_roi_frms).intersection(
                                    self.light_on_dict[cue_light]["pre_window_frames"]
                                )
                            )
                            self.overlap_post_window_frames = list(
                                set(inside_roi_frms).intersection(
                                    self.light_on_dict[cue_light]["post_window_frames"]
                                )
                            )
                            for state_name, lst in zip(
                                ["pre-cue", "cue", "post-cue"],
                                [
                                    self.overlap_pre_window_frames,
                                    self.overlap_light,
                                    self.overlap_post_window_frames,
                                ],
                            ):
                                self.roi_results[self.video_name][animal_name][
                                    roi_name
                                ][cue_light][state_name] = round(
                                    ((len(lst) * (1000 / self.fps)) / 1000), 4
                                )

    def organize_results(self):
        """
        Method to organize movement results into dataframe

        Returns
        -------
        Attribute: pd.DataFrame
            results_roi_df
        """

        self.results_df = pd.DataFrame(
            columns=["Video", "Animal", "Cue light", "Time period", "Measure", "Value"]
        )
        for video_name, video_data in self.results.items():
            for animal_name, animal_data in video_data.items():
                for light_name, light_data in animal_data.items():
                    for period_name, period_data in light_data.items():
                        for measure_name, measure_data in period_data.items():
                            self.results_df.loc[len(self.results_df)] = [
                                video_name,
                                animal_name,
                                light_name,
                                period_name,
                                measure_name,
                                measure_data,
                            ]

        if self.roi_setting:
            self.results_roi_df = pd.DataFrame(
                columns=[
                    "Video",
                    "Animal",
                    "ROI Name",
                    "Cue light",
                    "Time period",
                    "Time in ROI (s)",
                ]
            )
            for video_name, video_data in self.roi_results.items():
                for animal_name, animal_data in video_data.items():
                    for roi_name, roi_data in animal_data.items():
                        for light_name, light_data in roi_data.items():
                            for period_name, period_data in light_data.items():
                                self.results_roi_df.loc[len(self.results_roi_df)] = [
                                    video_name,
                                    animal_name,
                                    roi_name,
                                    light_name,
                                    period_name,
                                    period_data,
                                ]

    def save_results(self):
        """
        Method to save movement cue light results into the SimBA project folder.
        Results are stored in the `project_folder/logs` directory of the SimBA project.

        Returns
        -------
        None
        """

        save_results_path = os.path.join(
            self.logs_path,
            "Cue_lights_movement_statistics_{}.csv".format(self.datetime),
        )
        self.results_df = self.results_df.sort_values("Video").reset_index(drop=True)
        self.results_df.to_csv(save_results_path)
        stdout_success(
            msg="Cue light movement statistics saved in project_folder/logs directory."
        )
        if self.roi_setting:
            save_roi_results_path = os.path.join(
                self.logs_path, "Cue_lights_roi_statistics_{}.csv".format(self.datetime)
            )
            self.results_roi_df = self.results_roi_df.sort_values("Video").reset_index(
                drop=True
            )
            self.results_roi_df.to_csv(save_roi_results_path)
            stdout_success(
                msg="Cue light ROI statistics saved in project_folder/logs directory."
            )


# test = CueLightMovementAnalyzer(config_path=r'/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/project_config.ini',
#                                 pre_window=100, post_window=100, cue_light_names=['Rectangl_1'], threshold=0.0, roi_setting=True)
#
#
# def __init__(self,
#              config_path: str,
#              pre_window: int,
#              post_window: int,
#              cue_light_names: list,
#              threshold: float,
#              roi_setting: bool):



__author__ = "Simon Nilsson"

from tkinter import *

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.pose_processors.pose_reset import PoseResetter
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        Entry_Box, FileSelect)
from simba.utils.checks import check_str
from simba.utils.enums import ConfigKey, Keys, Links
from simba.utils.printing import stdout_success, stdout_trash
from simba.utils.read_write import tabulate_clf_info


class AddClfPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, config_path=config_path, title="ADD CLASSIFIER")
        ConfigReader.__init__(self, config_path=config_path)
        self.clf_eb = Entry_Box(self.main_frm, "CLASSIFIER NAME", "15")
        add_btn = Button(
            self.main_frm, text="ADD CLASSIFIER", command=lambda: self.run()
        )
        self.clf_eb.grid(row=0, column=0, sticky=NW)
        add_btn.grid(row=1, column=0, sticky=NW)

    def run(self):
        clf_name = self.clf_eb.entry_get.strip()
        check_str(name="CLASSIFIER NAME", value=clf_name)
        self.config.set(
            ConfigKey.SML_SETTINGS.value,
            ConfigKey.TARGET_CNT.value,
            str(self.clf_cnt + 1),
        )
        self.config.set(
            ConfigKey.SML_SETTINGS.value, f"model_path_{str(self.clf_cnt + 1)}", ""
        )
        self.config.set(
            ConfigKey.SML_SETTINGS.value,
            f"target_name_{str(self.clf_cnt + 1)}",
            clf_name,
        )
        self.config.set(
            ConfigKey.THRESHOLD_SETTINGS.value,
            f"threshold_{str(self.clf_cnt + 1)}",
            "None",
        )
        self.config.set(
            ConfigKey.MIN_BOUT_LENGTH.value, f"min_bout_{str(self.clf_cnt + 1)}", "None"
        )
        with open(self.config_path, "w") as f:
            self.config.write(f)
        stdout_success(
            msg=f"{clf_name} classifier added to SimBA project",
            source=self.__class__.__name__,
        )


class RemoveAClassifierPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(self, title="Warning: Remove classifier(s) settings")
        ConfigReader.__init__(self, config_path=config_path)
        self.remove_clf_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="SELECT A CLASSIFIER TO REMOVE",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.REMOVE_CLF.value,
        )
        self.clf_dropdown = DropDownMenu(
            self.remove_clf_frm, "Classifier", self.clf_names, "12"
        )
        self.clf_dropdown.setChoices(self.clf_names[0])

        run_btn = Button(
            self.main_frm, text="REMOVE CLASSIFIER", command=lambda: self.run()
        )
        self.remove_clf_frm.grid(row=0, sticky=W)
        self.clf_dropdown.grid(row=0, sticky=W)
        run_btn.grid(row=1, pady=10)

    def run(self):
        for i in range(len(self.clf_names)):
            self.config.remove_option(
                "SML settings", "model_path_{}".format(str(i + 1))
            )
            self.config.remove_option(
                "SML settings", "target_name_{}".format(str(i + 1))
            )
            self.config.remove_option(
                "threshold_settings", "threshold_{}".format(str(i + 1))
            )
            self.config.remove_option(
                "Minimum_bout_lengths", "min_bout_{}".format(str(i + 1))
            )
        self.clf_names.remove(self.clf_dropdown.getChoices())
        self.config.set("SML settings", "no_targets", str(len(self.clf_names)))

        for clf_cnt, clf_name in enumerate(self.clf_names):
            self.config.set(
                "SML settings", "model_path_{}".format(str(clf_cnt + 1)), ""
            )
            self.config.set(
                "SML settings", "target_name_{}".format(str(clf_cnt + 1)), clf_name
            )
            self.config.set(
                "threshold_settings", "threshold_{}".format(str(clf_cnt + 1)), "None"
            )
            self.config.set(
                "Minimum_bout_lengths", "min_bout_{}".format(str(clf_cnt + 1)), "None"
            )

        with open(self.config_path, "w") as f:
            self.config.write(f)

        stdout_trash(
            msg=f"{self.clf_dropdown.getChoices()} classifier removed from SimBA project.",
            source=self.__class__.__name__,
        )


# _ = RemoveAClassifierPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/Two_animals_16bps/project_folder/project_config.ini')


class PrintModelInfoPopUp(PopUpMixin):
    def __init__(self):
        PopUpMixin.__init__(self, title="PRINT MACHINE MODEL INFO", size=(250, 250))
        model_info_frame = LabelFrame(
            self.main_frm, text="PRINT MODEL INFORMATION", padx=5, pady=5, font="bold"
        )
        model_path_selector = FileSelect(
            model_info_frame, "Model path", title="Select a video file"
        )
        btn_print_info = Button(
            model_info_frame,
            text="PRINT MODEL INFO",
            command=lambda: tabulate_clf_info(clf_path=model_path_selector.file_path),
        )
        model_info_frame.grid(row=0, sticky=W)
        model_path_selector.grid(row=0, sticky=W, pady=5)
        btn_print_info.grid(row=1, sticky=W)


class PoseResetterPopUp(PopUpMixin):
    def __init__(self):
        PopUpMixin.__init__(self, title="WARNING!", size=(300, 100))
        popupframe = LabelFrame(self.main_frm)
        label = Label(
            popupframe, text="Do you want to remove user-defined pose-configurations?"
        )
        label.grid(row=0, columnspan=2)
        B1 = Button(
            popupframe,
            text="YES",
            fg="blue",
            command=lambda: PoseResetter(master=self.main_frm),
        )
        B2 = Button(popupframe, text="NO", fg="red", command=self.main_frm.destroy)
        popupframe.grid(row=0, columnspan=2)
        B1.grid(row=1, column=0, sticky=W)
        B2.grid(row=1, column=1, sticky=W)
        self.main_frm.mainloop()



import numpy as np
import pandas as pd
from scipy import stats
from scipy.signal import find_peaks
from statsmodels.stats.diagnostic import lilliefors


def rolling_frequentist_distribution_tests(data: np.array, feature_name: str, fps: int):
    """
    Helper to compare feature value distributions in 1s sequential time-bins: Kolmogorov-Smirnov and T-tests
    Compares the feature values against a normal distribution: Lillefors, Shapiro.
    Find the number of peaks in *rolling* 1s long feature window.
    """
    (ks_results,) = (np.full((data.shape[0]), -1.0),)
    t_test_results = np.full((data.shape[0]), -1.0)
    lillefors_results = np.full((data.shape[0]), -1.0)
    shapiro_results = np.full((data.shape[0]), -1.0)
    peak_cnt_results = np.full((data.shape[0]), -1.0)

    for i in range(fps, data.shape[0] - fps, fps):
        bin_1_idx, bin_2_idx = [i - fps, i], [i, i + fps]
        bin_1_data, bin_2_data = (
            data[bin_1_idx[0] : bin_1_idx[1]],
            data[bin_2_idx[0] : bin_2_idx[1]],
        )
        ks_results[i : i + fps + 1] = stats.ks_2samp(
            data1=bin_1_data, data2=bin_2_data
        ).statistic
        t_test_results[i : i + fps + 1] = stats.ttest_ind(
            bin_1_data, bin_2_data
        ).statistic

    for i in range(0, data.shape[0] - fps, fps):
        lillefors_results[i : i + fps + 1] = lilliefors(data[i : i + fps])[0]
        shapiro_results[i : i + fps + 1] = stats.shapiro(data[i : i + fps])[0]

    rolling_idx = np.arange(fps)[None, :] + 1 * np.arange(data.shape[0])[:, None]
    for i in range(rolling_idx.shape[0]):
        bin_start_idx, bin_end_idx = rolling_idx[i][0], rolling_idx[i][-1]
        peaks, _ = find_peaks(data[bin_start_idx:bin_end_idx], height=0)
        peak_cnt_results[i] = len(peaks)

    columns = [
        f"{feature_name}_KS",
        f"{feature_name}_TTEST",
        f"{feature_name}_LILLEFORS",
        f"{feature_name}_SHAPIRO",
        f"{feature_name}_PEAK_CNT",
    ]
    return pd.DataFrame(
        np.column_stack(
            (
                ks_results,
                t_test_results,
                lillefors_results,
                shapiro_results,
                peak_cnt_results,
            )
        ),
        columns=columns,
    ).round(4)


# features = [f'Convex_hull_mean_25_window']
# DATA_PATH = '/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/csv/features_extracted/SF2.csv'
# data = pd.read_csv(DATA_PATH, index_col=0)
# for feature in features:
#     df = rolling_frequentist_distribution_tests(data=data[feature].values,
#                                                 feature_name=feature, fps=25)
#     df.to_csv('Test.csv')



__author__ = "Simon Nilsson"

import os
from copy import deepcopy
from typing import Union

import cv2
import numpy as np

from simba.utils.read_write import get_fn_ext, get_video_meta_data


class CalculatePixelDistanceTool(object):
    def __init__(
        self, video_path: Union[str, os.PathLike], known_mm_distance: float
    ) -> None:
        self.video_path = video_path
        self.known_mm_distance = known_mm_distance
        self.video_dir, self.video_name, self.video_ext = get_fn_ext(video_path)
        if not os.access(video_path, os.R_OK):
            print("{} is not readable.".format(video_path))
            raise FileNotFoundError

        self.video_meta_data = get_video_meta_data(self.video_path)
        self.cap = cv2.VideoCapture(video_path)
        self.cap.set(1, 0)
        self.sensitivity = 20
        _, self.frame = self.cap.read()
        self.original_img = deepcopy(self.frame)
        max_res = max(self.video_meta_data["width"], self.video_meta_data["height"])
        self.add_spacer = 2
        space_scaler, radius_scaler, resolution_scaler, font_scaler = 80, 20, 1500, 1.5
        self.circle_scale = int(radius_scaler / (resolution_scaler / max_res))
        self.font_scale = float(font_scaler / (resolution_scaler / max_res))
        self.spacing_scale = int(font_scaler / (resolution_scaler / max_res))
        self.coordinate_lst = []
        self.choose_first_coordinate()
        self.choose_second_coordinate()
        self.manipulate_choices()

    def create_first_instructions_window(self):
        self.side_img = np.ones(
            (int(self.video_meta_data["height"] / 2), self.video_meta_data["width"], 3)
        )
        cv2.putText(
            self.side_img,
            "Double left mouse click on the first coordinate location.",
            (10, self.spacing_scale + 50),
            cv2.FONT_HERSHEY_SIMPLEX,
            self.font_scale,
            (10, 255, 10),
            3,
        )
        self.first_frame = np.uint8(np.concatenate((self.frame, self.side_img), axis=0))

    def create_second_instructions_window(self):
        self.side_img = np.ones(
            (int(self.video_meta_data["height"] / 2), self.video_meta_data["width"], 3)
        )
        cv2.putText(
            self.side_img,
            "Double left mouse click on the second coordinate location.",
            (10, self.spacing_scale + 50),
            cv2.FONT_HERSHEY_SIMPLEX,
            self.font_scale,
            (10, 255, 10),
            3,
        )
        self.second_frame = np.uint8(
            np.concatenate((self.frame, self.side_img), axis=0)
        )

    def create_third_instructions_window(self):
        self.side_img = np.ones(
            (int(self.video_meta_data["height"] / 2), self.video_meta_data["width"], 3)
        )
        cv2.putText(
            self.side_img,
            "Are you happy with the displayed choice?.",
            (10, self.spacing_scale + 50),
            cv2.FONT_HERSHEY_SIMPLEX,
            self.font_scale,
            (10, 255, 10),
            3,
        )
        cv2.putText(
            self.side_img,
            "Press ESC to proceed with the displayed choice.",
            (10, int(self.spacing_scale + 50 * (self.add_spacer * 2))),
            cv2.FONT_HERSHEY_SIMPLEX,
            self.font_scale,
            (255, 255, 255),
            2,
        )
        cv2.putText(
            self.side_img,
            "Double left mouse click on circle to move circle.",
            (10, int(self.spacing_scale + 50 * (self.add_spacer * 3))),
            cv2.FONT_HERSHEY_SIMPLEX,
            self.font_scale,
            (255, 255, 255),
            2,
        )
        cv2.putText(
            self.side_img,
            "Double left mouse click on circle to place circle to new location.",
            (10, int(self.spacing_scale + 50 * (self.add_spacer * 4))),
            cv2.FONT_HERSHEY_SIMPLEX,
            self.font_scale,
            (255, 255, 255),
            2,
        )
        cv2.putText(
            self.side_img,
            "Current pixels per millimeter: {}.".format(str(self.ppm)),
            (10, int(self.spacing_scale + 50 * (self.add_spacer * 5))),
            cv2.FONT_HERSHEY_SIMPLEX,
            self.font_scale,
            (255, 255, 255),
            2,
        )
        self.third_frame = np.uint8(np.concatenate((self.frame, self.side_img), axis=0))

    def calc_px_per_mm(self):
        euclid_px_dist = np.sqrt(
            (self.coordinate_lst[0][0] - self.coordinate_lst[1][0]) ** 2
            + (self.coordinate_lst[0][1] - self.coordinate_lst[0][1]) ** 2
        )
        self.ppm = round(euclid_px_dist / self.known_mm_distance, 5)

    def draw_all(self):
        for c in self.coordinate_lst:
            cv2.circle(
                self.third_frame,
                c,
                self.circle_scale,
                (10, 255, 10),
                -1,
                lineType=cv2.LINE_AA,
            )
        cv2.line(
            self.third_frame,
            self.coordinate_lst[0],
            self.coordinate_lst[1],
            (10, 255, 10),
            int(self.circle_scale / 5),
        )

    def choose_first_coordinate(self):
        cv2.namedWindow(
            "SELECT_COORDINATES: double left mouse click at two locations. Press ESC when done",
            cv2.WINDOW_NORMAL,
        )

        def get_coordinate(event, x, y, flags, param):
            if event == cv2.EVENT_LBUTTONDBLCLK:
                self.coordinate_lst.append((x, y))

        self.create_first_instructions_window()
        while True:
            cv2.setMouseCallback(
                "SELECT_COORDINATES: double left mouse click at two locations. Press ESC when done",
                get_coordinate,
            )
            cv2.imshow(
                "SELECT_COORDINATES: double left mouse click at two locations. Press ESC when done",
                self.first_frame,
            )
            k = cv2.waitKey(20) & 0xFF
            if (k == 27) or (len(self.coordinate_lst) == 1):
                cv2.destroyAllWindows()
                break

    def choose_second_coordinate(self):
        self.frame = deepcopy(self.original_img)
        cv2.namedWindow(
            "SELECT_COORDINATES: double left mouse click at two locations. Press ESC when done",
            cv2.WINDOW_NORMAL,
        )
        self.create_second_instructions_window()
        cv2.circle(
            self.second_frame,
            self.coordinate_lst[0],
            self.circle_scale,
            (10, 255, 10),
            -1,
            lineType=cv2.LINE_AA,
        )

        def get_coordinate(event, x, y, flags, param):
            if event == cv2.EVENT_LBUTTONDBLCLK:
                self.coordinate_lst.append((x, y))

        while True:
            cv2.setMouseCallback(
                "SELECT_COORDINATES: double left mouse click at two locations. Press ESC when done",
                get_coordinate,
            )
            cv2.imshow(
                "SELECT_COORDINATES: double left mouse click at two locations. Press ESC when done",
                self.second_frame,
            )
            k = cv2.waitKey(20) & 0xFF
            if (k == 27) or (len(self.coordinate_lst) == 2):
                cv2.destroyAllWindows()
                break

    def check_if_near_a_circle(self, clicked_cords):
        self.change_cord = None
        for cord_cnt, cord in enumerate(self.coordinate_lst):
            if (
                (clicked_cords[0] >= (cord[0] - self.sensitivity))
                and (clicked_cords[0] <= (cord[0] + self.sensitivity))
                and (clicked_cords[1] >= (cord[1] - self.sensitivity))
                and (clicked_cords[1] <= (cord[1] + self.sensitivity))
            ):
                self.change_cord = cord_cnt

    def manipulate_choices(self):
        self.frame = deepcopy(self.original_img)
        cv2.namedWindow(
            "SELECT_COORDINATES: double left mouse click at two locations. Press ESC when done",
            cv2.WINDOW_NORMAL,
        )
        self.calc_px_per_mm()
        self.create_third_instructions_window()
        self.draw_all()
        self.clicked_cord = (-np.inf, -np.inf)
        cv2.imshow(
            "SELECT_COORDINATES: double left mouse click at two locations. Press ESC when done",
            self.third_frame,
        )
        k = cv2.waitKey(20) & 0xFF

        def get_coordinate(event, x, y, flags, param):
            if event == cv2.EVENT_LBUTTONDBLCLK:
                self.clicked_cord = (x, y)

        def remove_single_coordinate():
            self.coordinate_lst.pop(self.change_cord)
            self.frame = deepcopy(self.original_img)
            cv2.circle(
                self.frame,
                self.coordinate_lst[0],
                self.circle_scale,
                (10, 255, 10),
                -1,
                lineType=cv2.LINE_AA,
            )
            cv2.namedWindow(
                "SELECT_COORDINATES: double left mouse click at two locations. Press ESC when done",
                cv2.WINDOW_NORMAL,
            )
            cv2.imshow(
                "SELECT_COORDINATES: double left mouse click at two locations. Press ESC when done",
                self.third_frame,
            )

        while True:
            cv2.setMouseCallback(
                "SELECT_COORDINATES: double left mouse click at two locations. Press ESC when done",
                get_coordinate,
            )
            cv2.imshow(
                "SELECT_COORDINATES: double left mouse click at two locations. Press ESC when done",
                self.third_frame,
            )
            k = cv2.waitKey(20) & 0xFF
            self.check_if_near_a_circle(self.clicked_cord)
            if (k == 27) and self.change_cord is not None:
                print(self.clicked_cord, self.change_cord)
                cv2.destroyAllWindows()
                remove_single_coordinate()
                self.clicked_cord = None



import os
import time
from tkinter import *

import cv2
import numpy as np
from PIL import Image, ImageTk

from simba.enums import Formats
from simba.misc_tools import (find_all_videos_in_directory, get_fn_ext,
                              get_video_meta_data)
from simba.mixins.config_reader import ConfigReader
from simba.utils.printing import stdout_success

MAX_FRM_SIZE = 1080, 650


class VideoRotator(ConfigReader):
    def __init__(self, config_path: str, input_path: str, output_dir: str):
        super().__init__(config_path=config_path)
        self.save_dir = output_dir
        if os.path.isfile(input_path):
            self.video_paths = [input_path]
        else:
            self.video_paths = find_all_videos_in_directory(directory=input_path)

    def __insert_img(self, img: np.array):
        current_frm_pil = Image.fromarray(img)
        current_frm_pil.thumbnail(MAX_FRM_SIZE, Image.ANTIALIAS)
        current_frm_pil = ImageTk.PhotoImage(
            master=self.main_frm, image=current_frm_pil
        )
        self.video_frame = Label(self.main_frm, image=current_frm_pil)
        self.video_frame.image = current_frm_pil
        self.video_frame.grid(row=0, column=0)

    def __rotate(self, value: int, img: np.array):
        self.dif_angle += value
        rotation_matrix = cv2.getRotationMatrix2D(
            (self.video_meta_data["width"] / 2, self.video_meta_data["height"] / 2),
            self.dif_angle,
            1,
        )
        img = cv2.warpAffine(
            img,
            rotation_matrix,
            (self.video_meta_data["width"], self.video_meta_data["height"]),
        )
        self.__insert_img(img=img)

    def __run_rotation(self):
        self.main_frm.destroy()
        fourcc = cv2.VideoWriter_fourcc(*Formats.MP4_CODEC.value)
        start = time.time()
        for video_cnt, (video_path, rotation) in enumerate(self.results.items()):
            cap = cv2.VideoCapture(video_path)
            rotation_matrix = cv2.getRotationMatrix2D(
                (self.video_meta_data["width"] / 2, self.video_meta_data["height"] / 2),
                rotation,
                1,
            )
            save_path = os.path.join(self.save_dir, os.path.basename(video_path))
            video_meta = get_video_meta_data(video_path=video_path)
            writer = cv2.VideoWriter(
                save_path,
                fourcc,
                video_meta["fps"],
                (video_meta["width"], video_meta["height"]),
            )
            img_cnt = 0
            while True:
                ret, img = cap.read()
                if not ret:
                    break
                img = cv2.warpAffine(
                    img,
                    rotation_matrix,
                    (self.video_meta_data["width"], self.video_meta_data["height"]),
                )
                writer.write(img)
                img_cnt += 1
                print(
                    f'Rotating frame {img_cnt}/{video_meta["frame_count"]} (Video {video_cnt+1}/{len(self.results.keys())}) '
                )
            cap.release()
            writer.release()
        stdout_success(
            msg=f"All videos rotated and saved in {self.save_dir}",
            elapsed_time=str(round((time.time() - start), 2)),
        )

    def __save(self):
        self.results[self.file_path] = self.dif_angle
        if len(self.results.keys()) == len(self.video_paths):
            self.__run_rotation()
        else:
            self.__run_interface(
                file_path=self.video_paths[len(self.results.keys()) - 1]
            )

    def __bind_keys(self):
        self.main_frm.bind(
            "<Left>", lambda x: self.__rotate(value=1, img=self._orig_img)
        )
        self.main_frm.bind(
            "<Right>", lambda x: self.__rotate(value=-1, img=self._orig_img)
        )
        self.main_frm.bind("<Escape>", lambda x: self.__save())

    def __run_interface(self, file_path: str):
        self.dif_angle = 0
        self.video_meta_data = get_video_meta_data(video_path=file_path)
        self.file_path = file_path
        _, self.video_name, _ = get_fn_ext(filepath=file_path)
        self.main_frm = Toplevel()
        self.main_frm.title(f"ROTATE VIDEO {self.video_name}")
        self.video_frm = Frame(self.main_frm)
        self.video_frm.grid(row=0, column=0)
        self.instruction_frm = Frame(self.main_frm, width=100, height=100)
        self.instruction_frm.grid(row=0, column=2, sticky=NW)
        self.key_lbls = Label(
            self.instruction_frm,
            text="\n\n Navigation: "
            "\n Left arrow = 1° left"
            "\n Right arrow = 1° right"
            "\n Esc = Save",
        )

        self.key_lbls.grid(sticky=NW)
        self.cap = cv2.VideoCapture(file_path)
        _, self.img = self.cap.read()
        self._orig_img = cv2.cvtColor(self.img, cv2.COLOR_RGB2BGR)
        self.__insert_img(img=self._orig_img)
        self.__bind_keys()

    def run(self):
        self.results = {}
        self.__run_interface(self.video_paths[0])
        # self.main_frm.mainloop()


rotator = VideoRotator(
    input_path="/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/videos/Together_1_downsampled.mp4",
    config_path="/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini",
    output_dir="/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/rotated",
)
rotator.run()



import re

import numpy as np


def update_all_tags(closest_roi):
    if closest_roi["Shape_type"] == "Rectangle":
        closest_roi["Tags"]["Center tag"] = (
            int(closest_roi["topLeftX"] + closest_roi["width"] / 2),
            int(closest_roi["topLeftY"] + closest_roi["height"] / 2),
        )
        closest_roi["Tags"]["Top left tag"] = (
            int(closest_roi["topLeftX"]),
            int(closest_roi["topLeftY"]),
        )
        closest_roi["Tags"]["Bottom right tag"] = (
            int(closest_roi["topLeftX"] + closest_roi["width"]),
            int(closest_roi["topLeftY"] + closest_roi["height"]),
        )
        closest_roi["Tags"]["Top right tag"] = (
            int(closest_roi["topLeftX"] + closest_roi["width"]),
            int(closest_roi["topLeftY"]),
        )
        closest_roi["Tags"]["Bottom left tag"] = (
            int(closest_roi["topLeftX"]),
            int(closest_roi["topLeftY"] + closest_roi["height"]),
        )
        closest_roi["Tags"]["Top tag"] = (
            int(closest_roi["topLeftX"] + closest_roi["width"] / 2),
            int(closest_roi["topLeftY"]),
        )
        closest_roi["Tags"]["Right tag"] = (
            int(closest_roi["topLeftX"] + closest_roi["width"]),
            int(closest_roi["topLeftY"] + closest_roi["height"] / 2),
        )
        closest_roi["Tags"]["Left tag"] = (
            int(closest_roi["topLeftX"]),
            int(closest_roi["topLeftY"] + closest_roi["height"] / 2),
        )
        closest_roi["Tags"]["Bottom tag"] = (
            int(closest_roi["topLeftX"] + closest_roi["width"] / 2),
            int(closest_roi["topLeftY"] + closest_roi["height"]),
        )

    elif closest_roi["Shape_type"] == "Circle":
        closest_roi["Tags"]["Center tag"] = (
            closest_roi["centerX"],
            closest_roi["centerY"],
        )
        closest_roi["Tags"]["Border tag"] = (
            int(closest_roi["centerX"] - closest_roi["radius"]),
            closest_roi["centerY"],
        )

    else:
        pass

    return closest_roi


def move_edge(closest_roi, closest_tag, new_click_loc):
    def rectangle_integrity_check(closest_roi):
        if closest_roi["topLeftX"] > closest_roi["Bottom_right_X"]:
            top_left_x, bottom_right_x = (
                closest_roi["topLeftX"],
                closest_roi["Bottom_right_X"],
            )
            closest_roi["topLeftX"] = bottom_right_x
            closest_roi["Bottom_right_X"] = top_left_x
            closest_roi["width"] = (
                closest_roi["Bottom_right_X"] - closest_roi["topLeftX"]
            )
            closest_roi["height"] = (
                closest_roi["Bottom_right_Y"] - closest_roi["topLeftY"]
            )

        if closest_roi["topLeftY"] > closest_roi["Bottom_right_Y"]:
            top_left_y, bottom_right_y = (
                closest_roi["topLeftY"],
                closest_roi["Bottom_right_Y"],
            )
            closest_roi["topLeftY"] = bottom_right_y
            closest_roi["Bottom_right_Y"] = top_left_y
            closest_roi["width"] = (
                closest_roi["Bottom_right_X"] - closest_roi["topLeftX"]
            )
            closest_roi["height"] = (
                closest_roi["Bottom_right_Y"] - closest_roi["topLeftY"]
            )

        return closest_roi

    if closest_roi["Shape_type"] == "Polygon":
        if closest_tag == "Center_tag":
            delta_x, delta_y = (
                closest_roi["Center_X"] - new_click_loc[0],
                closest_roi["Center_Y"] - new_click_loc[1],
            )
            closest_roi["Center_X"], closest_roi["Center_Y"] = (
                new_click_loc[0],
                new_click_loc[1],
            )
            new_array = np.zeros((0))
            for v in closest_roi["vertices"]:
                new_x = v[0] - delta_x
                new_y = v[1] - delta_y
                new_array = np.concatenate(
                    (new_array, np.array([new_x, new_y])), axis=0
                ).astype("int32")
            closest_roi["vertices"] = np.reshape(new_array, (-1, 2))
            polygon_pts_dict = {}
            for v, p in enumerate(closest_roi["vertices"]):
                polygon_pts_dict["Tag_" + str(v)] = (p[0], p[1])
            polygon_pts_dict["Center_tag"] = (
                closest_roi["Center_X"],
                int(closest_roi["Center_Y"]),
            )
            closest_roi["Tags"] = polygon_pts_dict

        else:
            tag_ix = int(re.sub("[^0-9]", "", closest_tag))
            closest_roi["vertices"][tag_ix] = new_click_loc
            closest_roi["Tags"][closest_tag] = new_click_loc
            poly_center = closest_roi["vertices"].mean(axis=0)
            closest_roi["Tags"]["Center_tag"] = (
                int(poly_center[0]),
                int(poly_center[1]),
            )
            closest_roi["Center_X"], closest_roi["Center_Y"] = int(poly_center[0]), int(
                poly_center[1]
            )

    elif closest_roi["Shape_type"] == "Rectangle":
        if closest_tag == "Right tag":
            closest_roi["Bottom_right_X"] = new_click_loc[0]
            closest_roi["width"] = (
                closest_roi["Bottom_right_X"] - closest_roi["topLeftX"]
            )
        if closest_tag == "Left tag":
            closest_roi["topLeftX"] = new_click_loc[0]
            closest_roi["width"] = (
                closest_roi["Bottom_right_X"] - closest_roi["topLeftX"]
            )
        if closest_tag == "Top tag":
            closest_roi["topLeftY"] = new_click_loc[1]
            closest_roi["height"] = (
                closest_roi["Bottom_right_Y"] - closest_roi["topLeftY"]
            )
        if closest_tag == "Bottom tag":
            closest_roi["Bottom_right_Y"] = new_click_loc[1]
            closest_roi["height"] = (
                closest_roi["Bottom_right_Y"] - closest_roi["topLeftY"]
            )
        if closest_tag == "Top left tag":
            closest_roi["topLeftX"], closest_roi["topLeftY"] = (
                new_click_loc[0],
                new_click_loc[1],
            )
            closest_roi["width"] = (
                closest_roi["Bottom_right_X"] - closest_roi["topLeftX"]
            )
            closest_roi["height"] = (
                closest_roi["Bottom_right_Y"] - closest_roi["topLeftY"]
            )
        if closest_tag == "Top right tag":
            closest_roi["Bottom_right_X"], closest_roi["topLeftY"] = (
                new_click_loc[0],
                new_click_loc[1],
            )
            closest_roi["width"] = (
                closest_roi["Bottom_right_X"] - closest_roi["topLeftX"]
            )
            closest_roi["height"] = (
                closest_roi["Bottom_right_Y"] - closest_roi["topLeftY"]
            )
        if closest_tag == "Bottom left tag":
            closest_roi["topLeftX"], closest_roi["Bottom_right_Y"] = (
                new_click_loc[0],
                new_click_loc[1],
            )
            closest_roi["width"] = (
                closest_roi["Bottom_right_X"] - closest_roi["topLeftX"]
            )
            closest_roi["height"] = (
                closest_roi["Bottom_right_Y"] - closest_roi["topLeftY"]
            )
        if closest_tag == "Bottom right tag":
            closest_roi["Bottom_right_X"], closest_roi["Bottom_right_Y"] = (
                new_click_loc[0],
                new_click_loc[1],
            )
            closest_roi["width"] = (
                closest_roi["Bottom_right_X"] - closest_roi["topLeftX"]
            )
            closest_roi["height"] = (
                closest_roi["Bottom_right_Y"] - closest_roi["topLeftY"]
            )
        if closest_tag == "Center tag":
            delta_x, delta_y = (
                closest_roi["Tags"]["Center tag"][0] - new_click_loc[0],
                closest_roi["Tags"]["Center tag"][1] - new_click_loc[1],
            )
            closest_roi["topLeftX"] = closest_roi["topLeftX"] - delta_x
            closest_roi["topLeftY"] = closest_roi["topLeftY"] - delta_y
            closest_roi["Bottom_right_X"] = (
                closest_roi["topLeftX"] + closest_roi["width"]
            )
            closest_roi["Bottom_right_Y"] = (
                closest_roi["topLeftY"] + closest_roi["height"]
            )

        closest_roi = rectangle_integrity_check(closest_roi)

        update_all_tags(closest_roi)

    elif closest_roi["Shape_type"] == "Circle":
        if closest_tag == "Center tag":
            closest_roi["centerX"], closest_roi["centerY"] = (
                new_click_loc[0],
                new_click_loc[1],
            )
        if closest_tag == "Border tag":
            closest_roi["radius"] = int(
                (
                    np.sqrt(
                        (closest_roi["centerX"] - new_click_loc[0]) ** 2
                        + (closest_roi["centerY"] - new_click_loc[1]) ** 2
                    )
                )
            )
        update_all_tags(closest_roi)


def move_edge_align(move_roi, move_tag, target_roi, target_tag):
    move_cord = target_roi["Tags"][target_tag]
    if move_roi["Shape_type"] == "Rectangle":
        if move_tag == "Top left tag":
            move_roi["topLeftX"] = move_cord[0]
            move_roi["topLeftY"] = move_cord[1]
        if move_tag == "Top tag":
            move_roi["topLeftX"] = move_cord[0] - int(move_roi["width"] / 2)
            move_roi["topLeftY"] = move_cord[1]
        if move_tag == "Top right tag":
            move_roi["topLeftX"] = move_cord[0] - move_roi["width"]
            move_roi["topLeftY"] = move_cord[1]
        if move_tag == "Left tag":
            move_roi["topLeftX"] = move_cord[0]
            move_roi["topLeftY"] = move_cord[1] - int(move_roi["height"] / 2)
        if move_tag == "Bottom left tag":
            move_roi["topLeftX"] = move_cord[0]
            move_roi["topLeftY"] = move_cord[1] - int(move_roi["height"])
        if move_tag == "Bottom tag":
            move_roi["topLeftX"] = move_cord[0] - int(move_roi["width"] / 2)
            move_roi["topLeftY"] = move_cord[1] - int(move_roi["height"])
        if move_tag == "Bottom right tag":
            move_roi["topLeftX"] = move_cord[0] - int(move_roi["width"])
            move_roi["topLeftY"] = move_cord[1] - int(move_roi["height"])
        if move_tag == "Right tag":
            move_roi["topLeftX"] = move_cord[0] - int(move_roi["width"])
            move_roi["topLeftY"] = move_cord[1] - int(move_roi["height"] / 2)
        if move_tag == "Center tag":
            move_roi["topLeftX"] = move_cord[0] - int(move_roi["width"] / 2)
            move_roi["topLeftY"] = move_cord[1] - int(move_roi["height"] / 2)

        move_roi["Bottom_right_X"] = move_roi["topLeftX"] + move_roi["width"]
        move_roi["Bottom_right_Y"] = move_roi["topLeftY"] + move_roi["height"]
        update_all_tags(move_roi)

    if move_roi["Shape_type"] == "Polygon":
        move_roi["Tags"][move_tag] = (move_cord[0], move_cord[1])
        new_array = np.zeros((0))
        for tag in list(move_roi["Tags"].keys())[:-1]:
            new_x = move_roi["Tags"][tag][0]
            new_y = move_roi["Tags"][tag][1]
            new_array = np.concatenate(
                (new_array, np.array([new_x, new_y])), axis=0
            ).astype("int32")
        move_roi["vertices"] = np.reshape(new_array, (-1, 2))
        poly_center = move_roi["vertices"].mean(axis=0)
        move_roi["Center_X"], move_roi["Center_Y"] = int(poly_center[0]), int(
            poly_center[1]
        )

    if move_roi["Shape_type"] == "Circle":
        if move_tag == "Center tag":
            move_roi["centerX"], move_roi["centerY"] = move_cord[0], move_cord[1]
        if move_tag == "Border tag":
            move_roi["centerX"], move_roi["centerY"] = (
                int(move_cord[0] + move_roi["radius"]),
                move_cord[1],
            )
        update_all_tags(move_roi)



__author__ = "Simon Nilsson"

from tkinter import *

from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.plotting.single_run_model_validation_video import \
    ValidateModelOneVideo
from simba.plotting.single_run_model_validation_video_mp import \
    ValidateModelOneVideoMultiprocess
from simba.ui.tkinter_functions import (CreateLabelFrameWithIcon, DropDownMenu,
                                        Entry_Box)
from simba.utils.checks import check_float, check_int
from simba.utils.enums import Formats, Keys, Links, Options
from simba.utils.read_write import check_file_exist_and_readable, str_2_bool


class ValidationVideoPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str, simba_main_frm: object):
        PopUpMixin.__init__(self, title="CREATE VALIDATION VIDEO")
        ConfigReader.__init__(self, config_path=config_path)
        self.feature_file_path = simba_main_frm.csvfile.file_path
        self.model_path = simba_main_frm.modelfile.file_path
        self.discrimination_threshold = simba_main_frm.dis_threshold.entry_get
        self.shortest_bout = simba_main_frm.min_behaviorbout.entry_get

        style_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="STYLE SETTINGS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.OUT_OF_SAMPLE_VALIDATION.value,
        )
        self.default_style_var = BooleanVar(value=True)
        default_style_cb = Checkbutton(
            style_frm,
            text="AUTO-COMPUTE STYLES",
            variable=self.default_style_var,
            command=lambda: self.enable_entrybox_from_checkbox(
                check_box_var=self.default_style_var,
                entry_boxes=[self.font_size_eb, self.spacing_eb, self.circle_size],
                reverse=True,
            ),
        )
        self.font_size_eb = Entry_Box(
            style_frm, "Font size: ", "25", validation="numeric"
        )
        self.spacing_eb = Entry_Box(
            style_frm, "Text spacing: ", "25", validation="numeric"
        )
        self.circle_size = Entry_Box(
            style_frm, "Circle size: ", "25", validation="numeric"
        )
        self.font_size_eb.entry_set(val=1)
        self.spacing_eb.entry_set(val=10)
        self.circle_size.entry_set(val=5)
        self.enable_entrybox_from_checkbox(
            check_box_var=self.default_style_var,
            entry_boxes=[self.font_size_eb, self.spacing_eb, self.circle_size],
            reverse=True,
        )

        tracking_frm = LabelFrame(
            self.main_frm,
            text="TRACKING OPTIONS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        )
        self.show_pose_dropdown = DropDownMenu(
            tracking_frm, "Show pose:", Options.BOOL_STR_OPTIONS.value, "20"
        )
        self.show_animal_names_dropdown = DropDownMenu(
            tracking_frm, "Show animal names:", Options.BOOL_STR_OPTIONS.value, "20"
        )
        self.show_pose_dropdown.setChoices(Options.BOOL_STR_OPTIONS.value[0])
        self.show_animal_names_dropdown.setChoices(Options.BOOL_STR_OPTIONS.value[1])

        multiprocess_frame = LabelFrame(
            self.main_frm,
            text="MULTI-PROCESS SETTINGS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        )
        self.multiprocess_var = BooleanVar(value=False)
        self.multiprocess_cb = Checkbutton(
            multiprocess_frame,
            text="Multiprocess videos (faster)",
            variable=self.multiprocess_var,
            command=lambda: self.enable_dropdown_from_checkbox(
                check_box_var=self.multiprocess_var,
                dropdown_menus=[self.multiprocess_dropdown],
            ),
        )
        self.multiprocess_dropdown = DropDownMenu(
            multiprocess_frame, "CPU cores:", list(range(2, self.cpu_cnt)), "12"
        )
        self.multiprocess_dropdown.setChoices(2)
        self.multiprocess_dropdown.disable()

        gantt_frm = LabelFrame(
            self.main_frm,
            text="GANTT SETTINGS",
            font=Formats.LABELFRAME_HEADER_FORMAT.value,
        )
        self.gantt_dropdown = DropDownMenu(
            gantt_frm, "GANTT TYPE:", Options.GANTT_VALIDATION_OPTIONS.value, "12"
        )
        self.gantt_dropdown.setChoices(Options.GANTT_VALIDATION_OPTIONS.value[0])

        style_frm.grid(row=0, column=0, sticky=NW)
        default_style_cb.grid(row=0, column=0, sticky=NW)
        self.font_size_eb.grid(row=1, column=0, sticky=NW)
        self.spacing_eb.grid(row=2, column=0, sticky=NW)
        self.circle_size.grid(row=3, column=0, sticky=NW)

        tracking_frm.grid(row=1, column=0, sticky=NW)
        self.show_pose_dropdown.grid(row=0, column=0, sticky=NW)
        self.show_animal_names_dropdown.grid(row=1, column=0, sticky=NW)

        multiprocess_frame.grid(row=2, column=0, sticky=NW)
        self.multiprocess_cb.grid(row=0, column=0, sticky=NW)
        self.multiprocess_dropdown.grid(row=0, column=1, sticky=NW)
        gantt_frm.grid(row=3, column=0, sticky=NW)
        self.gantt_dropdown.grid(row=0, column=0, sticky=NW)

        self.create_run_frm(run_function=self.run)
        self.main_frm.mainloop()

    def run(self):
        settings = {
            "pose": str_2_bool(self.show_pose_dropdown.getChoices()),
            "animal_names": str_2_bool(self.show_animal_names_dropdown.getChoices()),
        }
        settings["styles"] = None
        if not self.default_style_var.get():
            settings["styles"] = {}
            check_float(name="FONT SIZE", value=self.font_size_eb.entry_get)
            check_float(name="CIRCLE SIZE", value=self.circle_size.entry_get)
            check_float(name="SPACE SCALE", value=self.spacing_eb.entry_get)
            settings["styles"]["circle size"] = int(self.circle_size.entry_get)
            settings["styles"]["font size"] = int(self.font_size_eb.entry_get)
            settings["styles"]["space_scale"] = int(self.spacing_eb.entry_get)
        check_int(name="MINIMUM BOUT LENGTH", value=self.shortest_bout)
        check_float(
            name="DISCRIMINATION THRESHOLD", value=self.discrimination_threshold
        )
        check_file_exist_and_readable(file_path=self.feature_file_path)
        check_file_exist_and_readable(file_path=self.model_path)

        if not self.multiprocess_var.get():
            validation_video_creator = ValidateModelOneVideo(
                config_path=self.config_path,
                feature_file_path=self.feature_file_path,
                model_path=self.model_path,
                discrimination_threshold=float(self.discrimination_threshold),
                shortest_bout=int(self.shortest_bout),
                settings=settings,
                create_gantt=self.gantt_dropdown.getChoices(),
            )

        else:
            validation_video_creator = ValidateModelOneVideoMultiprocess(
                config_path=self.config_path,
                feature_file_path=self.feature_file_path,
                model_path=self.model_path,
                discrimination_threshold=float(self.discrimination_threshold),
                shortest_bout=int(self.shortest_bout),
                cores=int(self.multiprocess_dropdown.getChoices()),
                settings=settings,
                create_gantt=self.gantt_dropdown.getChoices(),
            )
        validation_video_creator.run()



__author__ = "Simon Nilsson"

import glob
import itertools
import os
import webbrowser
from tkinter import *
from typing import Union

import pandas as pd

from simba.cue_light_tools.cue_light_analyzer import CueLightAnalyzer
from simba.cue_light_tools.cue_light_clf_statistics import CueLightClfAnalyzer
from simba.cue_light_tools.cue_light_movement_statistics import \
    CueLightMovementAnalyzer
from simba.cue_light_tools.cue_light_visualizer import CueLightVisualizer
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import CreateLabelFrameWithIcon
from simba.utils.checks import check_float, check_int
from simba.utils.enums import Keys, Links
from simba.utils.errors import (CountError, NoChoosenClassifierError,
                                NoFilesFoundError, NoROIDataError)
from simba.utils.read_write import (find_video_of_file, get_all_clf_names,
                                    get_fn_ext, read_config_entry)


class CueLightAnalyzerMenu(ConfigReader, PopUpMixin):
    """
    Launch cue light analysis GUI in SimBA.

    :parameter str config_path: path to SimBA project config file in Configparser format

    .. note::
       `Cue light tutorials <https://github.com/sgoldenlab/simba/blob/master/docs/cue_light_tutorial.md>`__.


    Examples
    ----------
    >>> cue_light_gui = CueLightAnalyzerMenu(config_path='MySimBAConfigPath')
    >>> cue_light_gui.main_frm.mainloop()
    """

    def __init__(self, config_path: Union[str, os.PathLike]):
        ConfigReader.__init__(self, config_path=config_path)
        self.data_dir = os.path.join(
            self.project_path, "csv", "outlier_corrected_movement_location"
        )
        self.cue_light_data_folder = os.path.join(
            self.project_path, "csv", "cue_lights"
        )
        self.read_roi_data()
        if len(self.shape_names) == 0:
            raise CountError(
                msg="SIMBA ERROR: Cue light analysis require ROI definitions. Please define ROIs before doing cue light analysis"
            )
        PopUpMixin.__init__(self, size=(750, 300), title="SIMBA CUE LIGHT ANALYZER")
        self.lights_dict = {}

        self.cue_light_settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="DEFINE CUE LIGHTS",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.CUE_LIGHTS.value,
        )
        self.choose_lights_cnt_lbl = Label(
            self.cue_light_settings_frm, text="# Cue lights", width=10, anchor=W
        )
        self.choose_lights_cnt_var = IntVar()
        self.choose_lights_cnt_var.set(1)
        self.choose_lights_cnt_dropdown = OptionMenu(
            self.cue_light_settings_frm,
            self.choose_lights_cnt_var,
            *list(range(1, len(self.shape_names) + 1)),
            command=self.__create_cue_light_menus
        )
        self.cue_light_settings_frm.grid(row=0, sticky=NW)
        self.choose_lights_cnt_lbl.grid(row=0, column=0, sticky=W)
        self.choose_lights_cnt_dropdown.grid(row=0, column=1, sticky=W)
        self.analyze_data_frm = LabelFrame(
            self.main_frm,
            text="ANALYZE",
            font=("Helvetica", 15, "bold"),
            pady=5,
            padx=15,
        )
        self.analyze_cue_light_data_btn = Button(
            self.analyze_data_frm,
            text="Analyze cue light data",
            command=lambda: self.__analyze_cue_light_data(),
        )
        self.visualize_cue_light_data_btn = Button(
            self.analyze_data_frm,
            text="Visualize cue light data",
            command=lambda: self.__visualize_cue_light_data(),
        )
        self.video_var = BooleanVar()
        self.frames_var = BooleanVar()
        self.video_check = Checkbutton(
            self.analyze_data_frm, text="Create videos", variable=self.video_var
        )
        self.frames_check = Checkbutton(
            self.analyze_data_frm, text="Create frames", variable=self.frames_var
        )
        self.analyze_movements_btn = Button(
            self.analyze_data_frm,
            text="Analyze cue light movement",
            command=lambda: self.__inititate_animal_movement_menu(),
        )
        self.analyze_clf_btn = Button(
            self.analyze_data_frm,
            text="Analyze cue light classifications",
            command=lambda: self.__inititate_clf_menu(),
        )
        self.analyze_data_frm.grid(row=0, column=1, sticky=NW)
        self.analyze_cue_light_data_btn.grid(row=0, column=0, sticky=W)
        self.visualize_cue_light_data_btn.grid(row=1, column=0, sticky=W)
        self.video_check.grid(row=1, column=1, sticky=W)
        self.frames_check.grid(row=1, column=2, sticky=W)
        self.analyze_movements_btn.grid(row=2, column=0, sticky=W)
        self.analyze_clf_btn.grid(row=3, column=0, sticky=W)
        lbl_info_1 = Label(
            self.main_frm, text="NOTE: Cue light analysis require ROI definitions"
        )
        lbl_info_2 = Label(
            self.main_frm,
            text="[Click here to learn about cue light analysis]",
            cursor="hand2",
            fg="blue",
        )
        lbl_info_2.bind(
            "<Button-1>",
            lambda e: webbrowser.open_new(
                "https://github.com/sgoldenlab/simba/blob/master/docs/cue_light_tutorial.md"
            ),
        )
        lbl_info_1.grid(row=5, column=0, sticky=W)
        lbl_info_2.grid(row=6, column=0, sticky=W)

        mainloop()

    def __get_cue_light_names(self):
        self.light_lst = []
        for light_name, light_data in self.lights_dict.items():
            self.light_lst.append(light_data["light_chosen"].get())

    def __create_cue_light_menus(self, no_cue_lights):
        if len(self.lights_dict.keys()) > 0:
            for light_cnt in range(len(self.lights_dict.keys())):
                self.lights_dict[light_cnt]["label"].grid_forget()
                self.lights_dict[light_cnt]["dropdown"].grid_forget()
        for light_cnt in range(no_cue_lights):
            self.lights_dict[light_cnt] = {}
            current_row = 1 + light_cnt
            self.lights_dict[light_cnt]["label"] = Label(
                self.cue_light_settings_frm,
                text="Cue light {}".format(str(light_cnt + 1)),
                width=10,
                anchor=W,
            )
            self.lights_dict[light_cnt]["light_chosen"] = StringVar()
            self.lights_dict[light_cnt]["light_chosen"].set(self.shape_names[light_cnt])
            self.lights_dict[light_cnt]["dropdown"] = OptionMenu(
                self.cue_light_settings_frm,
                self.lights_dict[light_cnt]["light_chosen"],
                *self.shape_names,
                command=None
            )
            self.lights_dict[light_cnt]["label"].grid(
                row=current_row, column=0, sticky=W
            )
            self.lights_dict[light_cnt]["dropdown"].grid(
                row=current_row, column=1, sticky=W
            )

    def __analyze_cue_light_data(self):
        self.__get_cue_light_names()
        cue_light_analyzer = CueLightAnalyzer(
            config_path=self.config_path,
            in_dir=self.data_dir,
            cue_light_names=self.light_lst,
        )
        cue_light_analyzer.run()

    def __visualize_cue_light_data(self):
        self.cue_light_data_files = glob.glob(
            self.cue_light_data_folder + "/*" + self.file_type
        )
        if len(self.cue_light_data_files) == 0:
            raise NoFilesFoundError(
                msg="SIMBA ERROR: Zero data files found. Please analyze cue light data prior to visualizing cue light data"
            )
        else:
            self.__get_cue_light_names()
            for data_path in self.cue_light_data_files:
                _, file_name, _ = get_fn_ext(data_path)
                video_path = find_video_of_file(self.video_dir, file_name)
                cue_light_visualizer = CueLightVisualizer(
                    config_path=self.config_path,
                    cue_light_names=self.light_lst,
                    video_path=video_path,
                    video_setting=self.video_var.get(),
                    frame_setting=self.frames_var.get(),
                )
                cue_light_visualizer.run()

    def __inititate_animal_movement_menu(self):
        self.movement_main_frame = Toplevel()
        self.movement_main_frame.minsize(400, 400)
        self.movement_main_frame.wm_title("SIMBA CUE LIGHT ANALYZER: MOVEMENTS")
        self.movement_main_frame.lift()
        self.animal_cnt_frm = LabelFrame(
            self.movement_main_frame,
            text="SETTINGS",
            font=("Helvetica", 15, "bold"),
            pady=5,
            padx=15,
        )
        self.choose_animal_cnt_lbl = Label(
            self.animal_cnt_frm, text="# Animals", width=10, anchor=W
        )
        self.choose_animal_cnt_var = IntVar()
        self.choose_animal_cnt_var.set(1)
        self.choose_animal_cnt_dropdown = OptionMenu(
            self.animal_cnt_frm,
            self.choose_animal_cnt_var,
            *list(range(1, self.animal_cnt + 1)),
            command=self.__create_animal_bp_menues
        )
        self.animal_cnt_frm.grid(row=0, column=0, sticky=W)
        self.choose_animal_cnt_lbl.grid(row=0, column=0, sticky=W)
        self.choose_animal_cnt_dropdown.grid(row=0, column=1)

    def __create_animal_bp_menues(self, no_animals):
        self.animal_dict = {}
        current_row = 0
        for animal_cnt in range(no_animals):
            self.animal_dict[animal_cnt] = {}
            current_row = 1 + animal_cnt
            self.animal_dict[animal_cnt]["label"] = Label(
                self.animal_cnt_frm,
                text="Animal {} body-part:".format(str(animal_cnt + 1)),
                width=17,
                anchor=W,
            )
            self.animal_dict[animal_cnt]["bp_chosen"] = StringVar()
            self.animal_dict[animal_cnt]["bp_chosen"].set(
                self.body_parts_lst[animal_cnt]
            )
            self.animal_dict[animal_cnt]["dropdown"] = OptionMenu(
                self.animal_cnt_frm,
                self.animal_dict[animal_cnt]["bp_chosen"],
                self.body_parts_lst,
                command=None,
            )
            self.animal_dict[animal_cnt]["label"].grid(
                row=current_row, column=0, sticky=W
            )
            self.animal_dict[animal_cnt]["dropdown"].grid(
                row=current_row, column=1, sticky=W
            )

        self.pre_window_var = IntVar()
        self.pre_window_var.set(0)
        self.pre_window_lbl = Label(
            self.animal_cnt_frm, text="Pre-cue window (ms)", width=17, anchor=W
        )
        self.pre_window_entry = Entry(
            self.animal_cnt_frm, width=6, textvariable=self.pre_window_var
        )
        self.post_window_var = IntVar()
        self.post_window_var.set(0)
        self.post_window_lbl = Label(
            self.animal_cnt_frm, text="Post-cue window (ms)", width=17, anchor=W
        )
        self.post_window_entry = Entry(
            self.animal_cnt_frm, width=6, textvariable=self.post_window_var
        )
        self.pre_window_lbl.grid(row=current_row + 1, column=0, sticky=W)
        self.pre_window_entry.grid(row=current_row + 1, column=1, sticky=W)
        self.post_window_lbl.grid(row=current_row + 2, column=0, sticky=W)
        self.post_window_entry.grid(row=current_row + 2, column=1, sticky=W)
        self.threshold_lbl = Label(
            self.animal_cnt_frm, text="Threshold (0.00 - 1.00)", width=17, anchor=W
        )
        self.threshold_var = IntVar()
        self.threshold_var.set(0.00)
        self.threshold_entry = Entry(
            self.animal_cnt_frm, width=6, textvariable=self.threshold_var
        )
        self.threshold_lbl.grid(row=current_row + 3, column=0, sticky=W)
        self.threshold_entry.grid(row=current_row + 3, column=1, sticky=W)
        self.roi_var = BooleanVar()
        self.roi_check = Checkbutton(
            self.animal_cnt_frm, text="Analyze ROI data", variable=self.roi_var
        )
        self.roi_check.grid(row=current_row + 4, column=0, sticky=W)
        self.analyze_movement_btn = Button(
            self.animal_cnt_frm,
            text="Analyze movement data",
            command=lambda: self.start_movement_analysis(),
        )
        self.analyze_movement_btn.grid(row=current_row + 5, column=0, sticky=W)

    def start_movement_analysis(self):
        self.__get_cue_light_names()
        if len(self.light_lst) == 0:
            raise CountError(
                msg="Zero user-specified cue lights. Please specify cue lights before analyzing movements"
            )
        if not self.config.has_section("Cue light analysis"):
            self.config.add_section("Cue light analysis")
        for animal_cnt, animal_data in self.animal_dict.items():
            self.config["Cue light analysis"][
                "animal_{}_bp".format(str(animal_cnt + 1))
            ] = self.animal_dict[animal_cnt]["bp_chosen"].get()
        with open(self.config_path, "w") as file:
            self.config.write(file)

        check_int("Pre-cue window", value=self.pre_window_var.get())
        check_int("Post-cue window", value=self.post_window_var.get())
        check_float("Threshold", value=self.threshold_entry.get())

        cue_light_movement_analyzer = CueLightMovementAnalyzer(
            config_path=self.config_path,
            pre_window=self.pre_window_var.get(),
            post_window=self.post_window_var.get(),
            cue_light_names=self.light_lst,
            threshold=float(self.threshold_entry.get()),
            roi_setting=self.roi_var.get(),
        )
        cue_light_movement_analyzer.calculate_whole_session_movement()
        cue_light_movement_analyzer.organize_results()
        cue_light_movement_analyzer.save_results()

    def __inititate_clf_menu(self):
        self.clf_main_frame = Toplevel()
        self.clf_main_frame.minsize(400, 400)
        self.clf_main_frame.wm_title("SIMBA CUE LIGHT ANALYZER: CLASSIFICATIONS")
        self.clf_main_frame.lift()
        self.clf_cnt = read_config_entry(
            self.config, "SML settings", "no_targets", "int"
        )
        self.clf_names = get_all_clf_names(self.config, self.clf_cnt)
        self.clf_frm = LabelFrame(
            self.clf_main_frame,
            text="SELECT CLASSIFIERS",
            font=("Helvetica", 15, "bold"),
            pady=5,
            padx=15,
        )
        self.clf_frm.grid(row=0, column=0, sticky=NW)
        self.__insert_clf_tick_boxes()
        self.window_size_frm = LabelFrame(
            self.clf_main_frame,
            text="WINDOW SIZE",
            font=("Helvetica", 15, "bold"),
            pady=5,
            padx=15,
        )
        self.window_size_frm.grid(row=1, column=0, sticky=NW)
        self.pre_window_var_clf = IntVar()
        self.pre_window_var_clf.set(0)
        self.pre_window_lbl = Label(
            self.window_size_frm, text="Pre-cue window (ms)", width=17, anchor=W
        )
        self.pre_window_entry = Entry(
            self.window_size_frm, width=6, textvariable=self.pre_window_var_clf
        )
        self.post_window_var_clf = IntVar()
        self.post_window_var_clf.set(0)
        self.post_window_lbl = Label(
            self.window_size_frm, text="Post-cue window (ms)", width=17, anchor=W
        )
        self.post_window_entry = Entry(
            self.window_size_frm, width=6, textvariable=self.post_window_var_clf
        )
        self.pre_window_lbl.grid(row=0, column=0, sticky=NW)
        self.pre_window_entry.grid(row=0, column=1, sticky=NW)
        self.post_window_lbl.grid(row=1, column=0, sticky=NW)
        self.post_window_entry.grid(row=1, column=1, sticky=NW)
        self.run_frm = LabelFrame(
            self.clf_main_frame,
            text="RUN",
            font=("Helvetica", 15, "bold"),
            pady=5,
            padx=15,
        )
        self.window_size_frm.grid(row=1, column=0, sticky=NW)
        self.run_frm.grid(row=2, column=0, sticky=NW)
        self.analyze_clf_btn = Button(
            self.run_frm,
            text="Analyze classifier data",
            command=lambda: self.__start_clf_analysis(),
        )
        self.analyze_clf_btn.grid(row=1, column=0, sticky=NW)

    def __insert_clf_tick_boxes(self):
        self.entry_dict = {}
        for clf_cnt, clf_name in enumerate(self.clf_names):
            self.entry_dict[clf_name] = {}
            self.entry_dict[clf_name]["var"] = BooleanVar()
            self.entry_dict[clf_name]["cb"] = Checkbutton(
                self.clf_frm, text=clf_name, variable=self.entry_dict[clf_name]["var"]
            )
            self.entry_dict[clf_name]["cb"].grid(row=clf_cnt, column=0, sticky=N)

    def __start_clf_analysis(self):
        check_int("Pre-cue window", value=self.pre_window_var_clf.get())
        check_int("Post-cue window", value=self.post_window_var_clf.get())
        clf_lst = []
        self.__get_cue_light_names()
        if len(self.light_lst) == 0:
            raise NoFilesFoundError(
                msg="SIMBA ERROR: Zero data files found. Please analyze cue light data prior to visualizing cue light data"
            )
        for clf_name, clf_setting in self.entry_dict.items():
            if clf_setting["var"].get():
                clf_lst.append(clf_name)
        if len(clf_lst) == 0:
            raise NoChoosenClassifierError()
        else:
            clf_cue_light_analyzer = CueLightClfAnalyzer(
                config_path=self.config_path,
                pre_window=self.pre_window_var_clf.get(),
                post_window=self.post_window_var_clf.get(),
                clf_list=clf_lst,
                cue_light_names=self.light_lst,
            )
            clf_cue_light_analyzer.analyze_clf()
            clf_cue_light_analyzer.organize_results()
            clf_cue_light_analyzer.save_data()


# test = CueLightAnalyzerMenu(config_path=r'/Users/simon/Desktop/envs/troubleshooting/naresh/project_folder/project_config.ini')
# test.main_frm.mainloop()



__author__ = "Simon Nilsson"

import os
from datetime import datetime
from typing import Dict, Union

import cv2
import numpy as np
import pandas as pd

from simba.mixins.config_reader import ConfigReader
from simba.mixins.feature_extraction_mixin import FeatureExtractionMixin
from simba.utils.checks import (check_if_filepath_list_is_empty,
                                check_that_column_exist)
from simba.utils.data import (create_color_palettes, detect_bouts, find_bins,
                              get_video_meta_data)
from simba.utils.errors import InvalidVideoFileError, NoDataError
from simba.utils.printing import SimbaTimer, stdout_success
from simba.utils.read_write import get_fn_ext, read_df
from simba.utils.warnings import NoDataFoundWarning


class SeverityFrameCalculator(ConfigReader, FeatureExtractionMixin):
    """
    Computes the "severity" of classification frame events based on how much
    the animals are moving. Frames are scored as less or more severe at lower and higher movements, respectively.

    :parameter str config_path: path to SimBA project config file in Configparser format.
    :parameter dict settings: how to calculate the severity. E.g., {'brackets': 10, 'clf': 'Attack', 'animals': ['Simon', 'JJ'], 'time': True, 'frames': False}.

    .. note::
       `Tutorial <https://github.com/sgoldenlab/simba/blob/master/docs/Scenario2.md>`__.

    Examples
    ----------
    >>> settings = {'brackets': 10, 'clf': 'Attack', 'animals': ['Simon', 'JJ'], 'time': True, 'frames': False, 'normalization': 'ALL VIDEOS', 'save_bin_definitions': True}
    >>> processor = SeverityFrameCalculator(config_path='project_folder/project_config.ini', settings=settings)
    >>> processor.run()
    >>> processor.save()
    """

    def __init__(self, config_path: Union[str, os.PathLike], settings: Dict):
        ConfigReader.__init__(self, config_path=config_path)
        self.settings = settings
        check_if_filepath_list_is_empty(
            filepaths=self.machine_results_paths,
            error_msg=f"SIMBA ERROR: Cannot process severity. {self.machine_results_dir} directory is empty",
        )
        save_name = os.path.join(
            f'severity_{datetime.now().strftime("%Y%m%d%H%M%S")}.csv'
        )
        self.movement_animal_bp_dict = {
            k: self.animal_bp_dict[k] for k in self.settings["animals"]
        }
        self.color_lst_lst = create_color_palettes(
            no_animals=len(list(self.animal_bp_dict.keys())),
            map_size=len(self.bp_headers),
        )
        definitions_save_name = os.path.join(
            f'severity_bin_definitions_{datetime.now().strftime("%Y%m%d%H%M%S")}.csv'
        )
        self.save_path = os.path.join(self.logs_path, save_name)
        self.definitions_path = os.path.join(self.logs_path, definitions_save_name)
        self.results = {}

    def __calculate_movements(self):
        self.movements = {}
        for file_cnt, file_path in enumerate(self.machine_results_paths):
            _, video_name, _ = get_fn_ext(file_path)
            print(
                f"Analyzing movements in {video_name} ({file_cnt+1}/{len(self.machine_results_paths)})..."
            )
            _, px_per_mm, fps = self.read_video_info(video_name=video_name)
            df = read_df(file_path=file_path, file_type=self.file_type)
            if self.settings["clf"] not in df.columns:
                NoDataFoundWarning(
                    msg=f'Skipping file {video_name} - {self.settings["clf"]} data not present in file'
                )
                continue
            video_movement = np.full((len(df)), 0)
            for animal_name, animal_bodyparts in self.movement_animal_bp_dict.items():
                animal_df = df[animal_bodyparts["X_bps"] + animal_bodyparts["Y_bps"]]
                animal_df = self.create_shifted_df(df=animal_df)
                for bp_x, bp_y in zip(
                    animal_bodyparts["X_bps"], animal_bodyparts["Y_bps"]
                ):
                    video_movement = np.add(
                        video_movement,
                        self.euclidean_distance(
                            animal_df[bp_x].values,
                            animal_df[f"{bp_x}_shifted"].values,
                            animal_df[bp_y].values,
                            animal_df[f"{bp_y}_shifted"].values,
                            px_per_mm,
                        ),
                    )
            self.movements[video_name] = video_movement.astype(np.int)

    def run(self):
        self.__calculate_movements()
        self.visualization_data = {}
        self.video_bins_info = find_bins(
            data=self.movements,
            bracket_type=self.settings["bracket_type"],
            bracket_cnt=self.settings["brackets"],
            normalization_method=self.settings["normalization"],
        )
        for file_cnt, file_path in enumerate(self.machine_results_paths):
            _, video_name, _ = get_fn_ext(file_path)
            print(
                f"Matching brackets to movements in video {video_name} ({file_cnt+1}/{len(self.machine_results_paths)})..."
            )
            self.results[video_name] = {}
            _, px_per_mm, fps = self.read_video_info(video_name=video_name)
            df = read_df(file_path=file_path, file_type=self.file_type).astype(int)
            check_that_column_exist(
                df=df, column_name=self.settings["clf"], file_name=file_path
            )
            move_df = pd.DataFrame(self.movements[video_name], columns=["MOVEMENT"])
            video_bins = np.array((0))
            video_bins = np.hstack(
                (video_bins, self.video_bins_info[video_name][:, -1])
            )
            hist = np.histogram([self.movements[video_name]], bins=video_bins)
            move_df["BIN"] = np.fmin(
                np.digitize(self.movements[video_name], hist[1]),
                self.settings["brackets"],
            )
            df = pd.concat([df, move_df], axis=1)
            if self.settings["visualize"]:
                self.visualization_data[video_name] = df
            clf_df = (
                df["BIN"][df[self.settings["clf"]] == 1]
                .astype(int)
                .reset_index(drop=True)
            )
            for i in range(0, self.settings["brackets"]):
                if self.settings["frames"]:
                    self.results[video_name][f"Grade {str(i + 1)} (frames)"] = len(
                        clf_df[clf_df == i]
                    )
                if self.settings["time"]:
                    self.results[video_name][f"Grade {str(i + 1)} (s)"] = round(
                        (len(clf_df[clf_df == i]) / fps), 4
                    )
        self.__save()
        if self.settings["visualize"]:
            self.__visualize()

    def __save(self):
        print("Saving data..")
        out_df = pd.DataFrame(columns=["VIDEO", "MEASUREMENT", "VALUE"])
        for video_name, video_data in self.results.items():
            for grade, grade_data in video_data.items():
                out_df.loc[len(out_df)] = [video_name, grade, grade_data]
        out_df.to_csv(self.save_path)
        self.timer.stop_timer()
        stdout_success(
            msg=f"Severity data saved at {self.save_path}",
            elapsed_time=self.timer.elapsed_time_str,
        )

        if self.settings["save_bin_definitions"]:
            results = pd.DataFrame(columns=["VIDEO", "SEVERITY_BIN", ">=", "<"])
            for video_name, video_bins in self.video_bins_info.items():
                for bin_cnt, video_bin in enumerate(video_bins):
                    results.loc[len(results)] = [
                        video_name,
                        bin_cnt + 1,
                        video_bin[0],
                        video_bin[1],
                    ]
            save_path = os.path.join(
                self.logs_path, f"severity_bin_definitions_{self.datetime}.csv"
            )
            results.to_csv(save_path)
            stdout_success(msg=f"Severity bracket definitions saved at {save_path}")

    def __visualize(self):
        bouts_df_lst = []
        video_timer, save_dir = SimbaTimer(start=True), os.path.join(
            self.project_path, "frames", "output", "severity_bouts"
        )
        if not os.path.exists(save_dir):
            os.makedirs(save_dir)
        for video_name, video_data in self.visualization_data.items():
            bouts_df = detect_bouts(
                data_df=video_data, target_lst=[self.settings["clf"]], fps=-1
            )
            bouts_df.insert(0, "VIDEO", video_name)
            bouts_df_lst.append(bouts_df)
        bouts_df = pd.concat(bouts_df_lst, axis=0)
        if self.settings["visualize_event_cnt"] == "ALL CLIPS":
            self.settings["visualize_event_cnt"] = len(bouts_df)
        elif self.settings["visualize_event_cnt"] > len(bouts_df):
            self.settings["visualize_event_cnt"] = len(bouts_df)
            print(
                f'User specified {self.settings["visualize_event_cnt"]} visualization but only {len(self.results)} bouts where found. Creating {len(self.results)} videos...'
            )
        bouts_df = bouts_df.sample(n=self.settings["visualize_event_cnt"]).reset_index(
            drop=True
        )
        for idx, r in bouts_df.iterrows():
            video_name, start_frm, end_frame = (
                r["VIDEO"],
                r["Start_frame"],
                r["End_frame"],
            )
            print(start_frm)
            bout_df = self.visualization_data[video_name].iloc[start_frm:end_frame]
            clip_path = os.path.join(
                save_dir,
                f'{video_name}_{self.settings["clf"]}_{start_frm}_{end_frame}.mp4',
            )
            video_path = self.find_video_of_file(
                video_dir=self.video_dir, filename=video_name
            )
            video_meta_data = get_video_meta_data(video_path)
            self.space_scale, self.radius_scale, self.res_scale, self.font_scale = (
                60,
                12,
                1500,
                1.1,
            )
            self.max_dim = max(video_meta_data["width"], video_meta_data["height"])
            self.circle_scale = int(self.radius_scale / (self.res_scale / self.max_dim))
            self.font_size = float(self.font_scale / (self.res_scale / self.max_dim))
            video_fps = int(video_meta_data["fps"] * self.settings["video_speed"])
            if video_fps < 1:
                video_fps = 1
            cap = cv2.VideoCapture(video_path)
            writer = cv2.VideoWriter(
                clip_path,
                cv2.VideoWriter_fourcc(*"mp4v"),
                video_fps,
                (int(video_meta_data["width"]), int(video_meta_data["height"])),
            )
            event_frm_count, frm_cnt, current_frm = (
                (end_frame - start_frm),
                0,
                start_frm,
            )
            cap.set(1, current_frm)

            while current_frm < end_frame:
                ret, img = cap.read()
                if self.settings["show_pose"]:
                    frm_pose = bout_df.loc[current_frm]
                    cv2.putText(
                        img,
                        f'Severity bin: {frm_pose["BIN"]}',
                        (10, 50),
                        cv2.FONT_HERSHEY_COMPLEX,
                        self.font_size,
                        (255, 0, 0),
                        2,
                    )
                    for animal_cnt, (animal_name, animal_body_parts) in enumerate(
                        self.animal_bp_dict.items()
                    ):
                        for bp_cnt, (x_name, y_name) in enumerate(
                            zip(animal_body_parts["X_bps"], animal_body_parts["Y_bps"])
                        ):
                            x = frm_pose[[x_name, y_name]].values
                            cv2.circle(
                                img,
                                (x[0], x[1]),
                                0,
                                self.color_lst_lst[animal_cnt][bp_cnt],
                                self.circle_scale,
                            )

                if not ret:
                    raise InvalidVideoFileError(
                        msg=f'Could not find frame {current_frm} in video {video_path}. Video {video_path} contains {video_meta_data["frame_count"]} frames.'
                    )
                writer.write(img)
                frm_cnt += 1
                print(
                    f"Frame {str(frm_cnt)} / {str(event_frm_count)}, Event {idx + 1}/{str(len(bouts_df))}, Video {video_name}"
                )
                current_frm += 1
            writer.release()
        video_timer.stop_timer()
        stdout_success(
            msg=f"Videos complete: saved in {save_dir}",
            elapsed_time=video_timer.elapsed_time_str,
        )


# settings = {'brackets': 10,
#             'clf': 'Attack',
#             'animals': ['Simon', 'JJ'],
#             'normalization': 'ALL VIDEOS', #BY VIDEO
#             'bracket_type': "QUANTILE",
#             'time': True,
#             'frames': True,
#             'save_bin_definitions': True,
#             'visualize': True,
#             'visualize_event_cnt': 'ALL CLIPS',
#             'video_speed': 0.1,
#             'show_pose': True}
# processor = SeverityFrameCalculator(config_path='/Users/simon/Desktop/envs/troubleshooting/two_black_animals_14bp/project_folder/project_config.ini', settings=settings)
# processor.run()



__author__ = "Simon Nilsson"

from tkinter import *

from simba.data_processors.kleinberg_calculator import KleinbergCalculator
from simba.mixins.config_reader import ConfigReader
from simba.mixins.pop_up_mixin import PopUpMixin
from simba.ui.tkinter_functions import CreateLabelFrameWithIcon, Entry_Box
from simba.utils.checks import check_float, check_int
from simba.utils.enums import Keys, Links
from simba.utils.errors import NoChoosenClassifierError


class KleinbergPopUp(PopUpMixin, ConfigReader):
    def __init__(self, config_path: str):
        PopUpMixin.__init__(
            self, title="APPLY KLEINBERG BEHAVIOR CLASSIFICATION SMOOTHING"
        )
        ConfigReader.__init__(self, config_path=config_path)
        kleinberg_settings_frm = CreateLabelFrameWithIcon(
            parent=self.main_frm,
            header="Kleinberg Settings",
            icon_name=Keys.DOCUMENTATION.value,
            icon_link=Links.KLEINBERG.value,
        )
        self.k_sigma = Entry_Box(kleinberg_settings_frm, "Sigma", "10")
        self.k_sigma.entry_set("2")
        self.k_gamma = Entry_Box(kleinberg_settings_frm, "Gamma", "10")
        self.k_gamma.entry_set("0.3")
        self.k_hierarchy = Entry_Box(kleinberg_settings_frm, "Hierarchy", "10")
        self.k_hierarchy.entry_set("1")
        self.h_search_lbl = Label(kleinberg_settings_frm, text="Hierarchical search: ")
        self.h_search_lbl_val = BooleanVar()
        self.h_search_lbl_val.set(False)
        self.h_search_lbl_val_cb = Checkbutton(
            kleinberg_settings_frm, variable=self.h_search_lbl_val
        )
        kleinberg_table_frame = LabelFrame(
            self.main_frm, text="Choose classifier(s) to apply Kleinberg smoothing"
        )
        clf_var_dict, clf_cb_dict = {}, {}
        for clf_cnt, clf in enumerate(self.clf_names):
            clf_var_dict[clf] = BooleanVar()
            clf_cb_dict[clf] = Checkbutton(
                kleinberg_table_frame, text=clf, variable=clf_var_dict[clf]
            )
            clf_cb_dict[clf].grid(row=clf_cnt, sticky=NW)

        run_kleinberg_btn = Button(
            self.main_frm,
            text="Apply Kleinberg Smoother",
            command=lambda: self.run_kleinberg(
                behaviors_dict=clf_var_dict,
                hierarchical_search=self.h_search_lbl_val.get(),
            ),
        )

        kleinberg_settings_frm.grid(row=0, sticky=W, padx=10)
        self.k_sigma.grid(row=0, sticky=W)
        self.k_gamma.grid(row=1, sticky=W)
        self.k_hierarchy.grid(row=2, sticky=W)
        self.h_search_lbl.grid(row=3, column=0, sticky=W)
        self.h_search_lbl_val_cb.grid(row=3, column=1, sticky=W)
        kleinberg_table_frame.grid(row=1, pady=10, padx=10)
        run_kleinberg_btn.grid(row=2)

    def run_kleinberg(self, behaviors_dict: dict, hierarchical_search: bool):
        targets = []
        for behaviour, behavior_val in behaviors_dict.items():
            if behavior_val.get():
                targets.append(behaviour)

        if len(targets) == 0:
            raise NoChoosenClassifierError(source=self.__class__.__name__)

        check_int(name="Hierarchy", value=self.k_hierarchy.entry_get)
        check_float(name="Sigma", value=self.k_sigma.entry_get)
        check_float(name="Gamma", value=self.k_gamma.entry_get)

        try:
            print(
                "Applying kleinberg hyperparameter Setting: Sigma: {}, Gamma: {}, Hierarchy: {}".format(
                    str(self.k_sigma.entry_get),
                    str(self.k_gamma.entry_get),
                    str(self.k_hierarchy.entry_get),
                )
            )
        except:
            print("Please insert accurate values for all hyperparameters.")

        kleinberg_analyzer = KleinbergCalculator(
            config_path=self.config_path,
            classifier_names=targets,
            sigma=self.k_sigma.entry_get,
            gamma=self.k_gamma.entry_get,
            hierarchy=self.k_hierarchy.entry_get,
            hierarchical_search=hierarchical_search,
        )
        kleinberg_analyzer.run()


# _ = KleinbergPopUp(config_path='/Users/simon/Desktop/envs/troubleshooting/Two_animals_16bps/project_folder/project_config.ini')



__author__ = "Simon Nilsson"

import glob
import os
import pickle
from datetime import datetime
from typing import List, Optional, Union

import numpy as np
import pandas as pd
from sklearn.feature_selection import VarianceThreshold

import simba

try:
    from typing import Literal
except:
    from typing_extensions import Literal

from sklearn.preprocessing import (MinMaxScaler, QuantileTransformer,
                                   StandardScaler)

from simba.unsupervised.enums import UMLOptions, Unsupervised
from simba.utils.checks import check_float
from simba.utils.enums import Paths
from simba.utils.errors import (DirectoryNotEmptyError, IntegerError,
                                InvalidFileTypeError, InvalidInputError,
                                MissingColumnsError, NoDataError,
                                NoFilesFoundError)
from simba.utils.printing import SimbaTimer


class UnsupervisedMixin(object):
    """
    Methods for unsupervised ML.
    """

    def __init__(self):

        self.datetime = datetime.now().strftime("%Y%m%d%H%M%S")
        self.timer = SimbaTimer(start=True)
        model_names_dir = os.path.join(
            os.path.dirname(simba.__file__), Paths.UNSUPERVISED_MODEL_NAMES.value
        )
        self.model_names = list(
            pd.read_parquet(model_names_dir)[Unsupervised.NAMES.value]
        )

    def read_pickle(self, data_path: Union[str, os.PathLike]) -> dict:
        """
        Read a single or directory of pickled objects. If directory, returns dict with numerical sequential integer keys for
        each object.

        :param str data_path: Pickled file path, or directory of pickled files.
        :returns dict

        :example:
        >>> data = read_pickle(data_path='/test/unsupervised/cluster_models')
        """

        if os.path.isdir(data_path):
            data = {}
            files_found = glob.glob(data_path + "/*.pickle")
            if len(files_found) == 0:
                raise NoFilesFoundError(
                    msg=f"SIMBA ERROR: Zero pickle files found in {data_path}.",
                    source=self.__class__.__name__,
                )
            for file_cnt, file_path in enumerate(files_found):
                with open(file_path, "rb") as f:
                    data[file_cnt] = pickle.load(f)
        if os.path.isfile(data_path):
            with open(data_path, "rb") as f:
                data = pickle.load(f)
        return data

    def write_pickle(self, data: dict, save_path: Union[str, os.PathLike]) -> None:
        """
        Write a single object as pickle.

        :param str data_path: Pickled file path.
        :param str save_path: Location of saved pickle.

        :example:
        >>> data = test.write_pickle(data= my_model, save_path='/test/unsupervised/cluster_models/My_model.pickle')
        """

        try:
            with open(save_path, "wb") as f:
                pickle.dump(data, f, protocol=pickle.HIGHEST_PROTOCOL)
        except Exception as e:
            print(e.args[0])
            raise InvalidFileTypeError(
                msg="Data could not be saved as a pickle.",
                source=self.__class__.__name__,
            )

    def check_umap_hyperparameters(self, hyper_parameters: dict):
        """
        Checks if umap embedder hyperparameters are valid.

        :param dict hyper_parameters: Dictionary holding human hyerparameters.
        :raises InvalidInputError: If any input is invalid

        :example:
        >>> check_umap_hyperparameters(hyper_parameters={'n_neighbors': 2, 'min_distance': 0.1, 'spread': 1})
        """

        for key in Unsupervised.HYPERPARAMETERS.value:
            if key not in hyper_parameters.keys():
                raise InvalidInputError(
                    msg=f"Hyperparameter dictionary is missing {key} entry.",
                    source=self.__class__.__name__,
                )
        for key in [
            Unsupervised.N_NEIGHBORS.value,
            Unsupervised.MIN_DISTANCE.value,
            Unsupervised.SPREAD.value,
        ]:
            if len(hyper_parameters[key]) == 0:
                raise InvalidInputError(
                    msg=f"Hyperparameter dictionary key {key} has 0 entries.",
                    source=self.__class__.__name__,
                )
        if (
            hyper_parameters[Unsupervised.SCALER.value]
            not in UMLOptions.SCALER_OPTIONS.value
        ):
            raise InvalidInputError(
                msg=f"Scaler {Unsupervised.SCALER.value} not supported. Opitions: {UMLOptions.SCALER_OPTIONS.value}",
                source=self.__class__.__name__,
            )
        check_float(
            "VARIANCE THRESHOLD", value=hyper_parameters[Unsupervised.VARIANCE.value]
        )

    def find_low_variance_fields(
        self, data: pd.DataFrame, variance: float
    ) -> List[str]:
        """
        Finds fields with variance below provided threshold.

        :param pd.DataFrame data: Dataframe with continoues numerical features.
        :param float variance: Variance threshold (0.0-1.0).
        :return List[str]:
        """
        feature_selector = VarianceThreshold(threshold=round((variance / 100), 2))
        feature_selector.fit(data)
        low_variance_fields = [
            c
            for c in data.columns
            if c not in data.columns[feature_selector.get_support()]
        ]
        if len(low_variance_fields) == len(data.columns):
            raise NoDataError(
                msg=f"All feature columns show a variance below the {str(variance)} threshold. Thus, no data remain for analysis.",
                source=self.find_low_variance_fields.__name__,
            )
        return low_variance_fields

    def drop_fields(self, data: pd.DataFrame, fields: List[str]) -> pd.DataFrame:
        """
        Drops specified fields in dataframe.
        :param pd.DataFrame: Data in pandas format.
        :param  List[str] fields: Columns to drop.
        :return pd.DataFrame
        """
        return data.drop(columns=fields)

    def define_scaler(
        self, scaler_name: Literal["MIN-MAX", "STANDARD", "QUANTILE"]
    ) -> Union[MinMaxScaler, StandardScaler, QuantileTransformer]:
        """
        Creates sklearn scaler object. See ``UMLOptions.SCALER_OPTIONS.value`` for accepted scalers.

        :example:
        >>> define_scaler(scaler_name='MIN-MAX')
        """

        if scaler_name not in UMLOptions.SCALER_OPTIONS.value:
            raise InvalidInputError(
                msg=f"Scaler {scaler_name} not supported. Options: {UMLOptions.SCALER_OPTIONS.value}",
                source=self.__class__.__name__,
            )
        if scaler_name == Unsupervised.MIN_MAX.value:
            return MinMaxScaler()
        elif scaler_name == Unsupervised.STANDARD.value:
            return StandardScaler()
        elif scaler_name == Unsupervised.QUANTILE.value:
            return QuantileTransformer()

    def scaler_transform(
        self,
        data: pd.DataFrame,
        scaler: Literal[MinMaxScaler or StandardScaler or QuantileTransformer],
    ) -> pd.DataFrame:
        """
        Helper to run transform using fitted scaler.

        :param pd.DataFrame data: Data to transform.
        :param scaler: fitted scaler.
        """

        return pd.DataFrame(scaler.transform(data), columns=data.columns).set_index(
            data.index
        )

    def scaler_inverse_transform(
        self,
        data: pd.DataFrame,
        scaler=MinMaxScaler or StandardScaler or QuantileTransformer,
    ):
        return pd.DataFrame(
            scaler.inverse_transform(data), columns=data.columns
        ).set_index(data.index)

    def check_expected_fields(self, data_fields: list, expected_fields: list):
        remaining_fields = [x for x in data_fields if x not in expected_fields]
        if len(remaining_fields) > 0:
            raise MissingColumnsError(
                f"The data contains {str(len(remaining_fields))} unexpected field(s): {str(remaining_fields)}",
                source=self.__class__.__name__,
            )

    def check_key_exist_in_object(self, object: dict, key: str):
        if key not in object.keys():
            raise InvalidInputError(
                msg=f"{key} does not exist in object", source=self.__class__.__name__
            )

    def get_cluster_cnt(
        self, data: np.array, clusterer_name: str, minimum_clusters: Optional[int] = 1
    ) -> int:
        """
        Helper to check the number of unique observations in array. E.g., check the number of unique clusters.

        :param np.array data: 1D numpy array with cluster assignments.
        :param str clusterer_name: Arbitrary name of the fitted model used to generate ``data``.
        :param Optional[int] minimum_clusters: Minimum number of clusters allowed.
        :raises IntegerError: If unique cluster count is less than ``minimum_clusters``.
        """

        cnt = np.unique(data).shape[0]
        if cnt < minimum_clusters:
            raise IntegerError(
                msg=f"Clustrer {clusterer_name} has {str(cnt)} clusters, but {str(minimum_clusters)} clusters is required for the operation.",
                source=self.__class__.__name__,
            )
        return cnt

    def check_that_directory_is_empty(self, directory: Union[str, os.PathLike]) -> None:
        """
        Checks if a directory is empty

        :param str directory: Directory to check.
        :raises DirectoryNotEmptyError: If ``directory`` contains files.
        """

        try:
            all_files_in_folder = [
                f for f in next(os.walk(directory))[2] if not f[0] == "."
            ]
        except StopIteration:
            return 0
        else:
            if len(all_files_in_folder) > 0:
                raise DirectoryNotEmptyError(
                    msg=f"The {directory} is not empty and contains {str(len(all_files_in_folder))} files. Use a directory that is empty.",
                    source=self.__class__.__name__,
                )
